from __future__ import annotations

import datetime
import json
import os
import random
import time
from argparse import ArgumentParser
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Dict, List

import netsquid as ns
from netqasm.lang.instr.flavour import NVFlavour, TrappedIonFlavour

from qoala.lang.ehi import UnitModule
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    ClassicalConnectionConfig,
    LatenciesConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchResult, ProgramBatch, ProgramInput
from qoala.runtime.statistics import SchedulerStatistics
from qoala.sim.build import build_network_from_config
from qoala.util.logging import LogManager
from qoala.util.runner import (
    AppResult,
    create_batch,
    run_two_node_app,
    run_two_node_app_separate_inputs,
)


def relative_to_cwd(file: str) -> str:
    return os.path.join(os.path.dirname(__file__), file)


def create_procnode_cfg(
    name: str, id: int, num_qubits: int, hardware: str
) -> ProcNodeConfig:
    if hardware == "generic":
        return ProcNodeConfig(
            node_name=name,
            node_id=id,
            topology=TopologyConfig.perfect_config_uniform_default_params(num_qubits),
            latencies=LatenciesConfig(qnos_instr_time=1000),
            ntf=NtfConfig.from_cls_name("GenericNtf"),
        )
    elif hardware == "nv":
        return ProcNodeConfig(
            node_name=name,
            node_id=id,
            topology=TopologyConfig.perfect_nv_default_params(num_qubits=num_qubits),
            latencies=LatenciesConfig(qnos_instr_time=1000),
            ntf=NtfConfig.from_cls_name("NvNtf"),
        )
    elif hardware == "tri":
        return ProcNodeConfig(
            node_name=name,
            node_id=id,
            topology=TopologyConfig.perfect_tri_default_params(num_qubits=num_qubits),
            latencies=LatenciesConfig(qnos_instr_time=1000),
            ntf=NtfConfig.from_cls_name("TrappedIonNtf"),
        )


def load_program(path: str, hardware: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()

    if hardware == "generic":
        flavour = None
    elif hardware == "nv":
        flavour = NVFlavour()
    elif hardware == "tri":
        flavour = TrappedIonFlavour()
    return QoalaParser(text, flavour=flavour).parse()


@dataclass
class TeleportResult:
    alice_results: BatchResult
    bob_results: BatchResult


def run_teleport(
    hardware: str, num_iterations: int, different_inputs: bool = False
) -> TeleportResult:
    ns.sim_reset()

    alice_id = 1
    bob_id = 0

    alice_node_cfg = create_procnode_cfg("alice", alice_id, 2, hardware)
    bob_node_cfg = create_procnode_cfg("bob", bob_id, 1, hardware)

    cconn = ClassicalConnectionConfig.from_nodes(alice_id, bob_id, 1e9)
    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[alice_node_cfg, bob_node_cfg], link_duration=1000
    )
    network_cfg.cconns = [cconn]

    alice_file = f"teleport_{hardware}_alice.iqoala"
    bob_file = f"teleport_{hardware}_bob.iqoala"
    alice_program = load_program(alice_file, hardware)
    bob_program = load_program(bob_file, hardware)

    if different_inputs:
        alice_inputs: List[ProgramInput] = []
        bob_inputs: List[ProgramInput] = []
        for _ in range(num_iterations):
            basis = random.randint(0, 5)
            alice_inputs.append(ProgramInput({"bob_id": bob_id, "state": basis}))
            bob_inputs.append(ProgramInput({"alice_id": alice_id, "state": basis}))

        app_result = run_two_node_app_separate_inputs(
            num_iterations=num_iterations,
            programs={"alice": alice_program, "bob": bob_program},
            program_inputs={"alice": alice_inputs, "bob": bob_inputs},
            network_cfg=network_cfg,
            linear=True,
        )
    else:
        # state = 5 -> teleport |1> state
        alice_input = ProgramInput({"bob_id": bob_id, "state": 5})
        # state = 4 -> measure in +Z, i.e. expect a "1" outcome
        bob_input = ProgramInput({"alice_id": alice_id, "state": 4})

        app_result = run_two_node_app(
            num_iterations=num_iterations,
            programs={"alice": alice_program, "bob": bob_program},
            program_inputs={"alice": alice_input, "bob": bob_input},
            network_cfg=network_cfg,
            linear=True,
        )

    alice_result = app_result.batch_results["alice"]
    bob_result = app_result.batch_results["bob"]

    return TeleportResult(alice_result, bob_result)


def teleport_different_inputs(hardware: str, num_iterations: int):
    result = run_teleport(hardware, num_iterations, different_inputs=True)

    program_results = result.bob_results.results
    outcomes = [result.values["outcome"] for result in program_results]
    # print(outcomes)
    assert all(outcome == 0 for outcome in outcomes)


@dataclass
class DataPoint:
    success: bool
    sim_duration: float


@dataclass
class DataMeta:
    timestamp: str
    num_iterations: int


@dataclass
class Data:
    meta: DataMeta
    data_points: List[DataPoint]


def run_hardware(hardware: str, num_iterations: int) -> DataPoint:
    start = time.time()

    success: bool
    try:
        teleport_different_inputs(hardware, num_iterations)
        success = True
    except AssertionError:
        success = False

    end = time.time()
    duration = round(end - start, 2)
    print(f"duration: {duration} s")

    return DataPoint(success, duration)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--num_iterations", "-n", type=int, required=True)

    args = parser.parse_args()
    num_iterations = args.num_iterations

    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    data_points: List[DataPoint] = []
    for hardware in ["generic", "nv", "tri"]:
        data_point = run_hardware(hardware, num_iterations)
        data_points.append(data_point)

    abs_dir = relative_to_cwd(f"data")
    Path(abs_dir).mkdir(parents=True, exist_ok=True)
    last_path = os.path.join(abs_dir, "LAST.json")
    timestamp_path = os.path.join(abs_dir, f"{timestamp}.json")

    meta = DataMeta(timestamp=timestamp, num_iterations=num_iterations)
    data = Data(meta=meta, data_points=data_points)
    json_data = asdict(data)

    with open(last_path, "w") as datafile:
        json.dump(json_data, datafile)
    with open(timestamp_path, "w") as datafile:
        json.dump(json_data, datafile)



import logging
from typing import Dict, Generator, List, Optional, Tuple

from netsquid.components.component import Component, Port
from netsquid.protocols import Protocol

from pydynaa import EventExpression
from qoala.runtime.message import Message
from qoala.util.logging import LogManager


class MessageBuffer:
    def __init__(self) -> None:
        self._messages: Dict[
            Tuple[int, int], List[Message]
        ] = {}  # (src PID, dst PID) -> message list

    def add_msg(self, msg: Message) -> None:
        if (msg.src_pid, msg.dst_pid) not in self._messages:
            self._messages[(msg.src_pid, msg.dst_pid)] = [msg]
        else:
            self._messages[(msg.src_pid, msg.dst_pid)].append(msg)

    def has_msg(self, src_pid: int, dst_pid: int) -> bool:
        if (src_pid, dst_pid) not in self._messages:
            return False
        return len(self._messages[(src_pid, dst_pid)]) > 0

    def has_any(self) -> bool:
        if len(self._messages) == 0:
            return False
        return any(len(buf) > 0 for buf in self._messages.values())

    def get_all(self) -> List[Tuple[int, int]]:
        # Does *NOT* pop messages.
        return [
            (src, dst) for ((src, dst), buf) in self._messages.items() if len(buf) > 0
        ]

    def count_all(self) -> int:
        return sum(len(buf) for buf in self._messages.values())

    def pop_msg(self, src_pid: int, dst_pid: int) -> Message:
        return self._messages[(src_pid, dst_pid)].pop(0)

    def pop_any(self) -> Message:
        for buf in self._messages.values():
            if len(buf) > 0:
                return buf.pop(0)
        raise RuntimeError

    def pop_all(self) -> List[Message]:
        messages = []
        for buf in self._messages.values():
            messages.extend(buf)
            buf.clear()
        return messages


class PortListener(Protocol):
    def __init__(self, port: Port, signal_label: str) -> None:
        self._buffer = MessageBuffer()
        self._port = port
        self._signal_label = signal_label
        self.add_signal(signal_label)

    @property
    def buffer(self) -> MessageBuffer:
        return self._buffer

    def run(self) -> Generator[EventExpression, None, None]:
        while True:
            # Wait for an event saying that there is new input.
            yield self.await_port_input(self._port)

            counter = 0
            # Read all inputs and count them.
            while True:
                input = self._port.rx_input()
                if input is None:
                    break
                for item in input.items:
                    self._buffer.add_msg(item)
                counter += 1
            # If there are n inputs, there have been n events, but we yielded only
            # on one of them so far. "Flush" these n-1 additional events:
            while counter > 1:
                yield self.await_port_input(self._port)
                counter -= 1

            # Only after having yielded on all current events, we can schedule a
            # notification event, so that its reactor can handle all inputs at once.
            self.send_signal(self._signal_label)


class ComponentProtocol(Protocol):
    def __init__(self, name: str, comp: Component) -> None:
        super().__init__(name)
        self._listeners: Dict[str, PortListener] = {}
        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({comp.name})"
        )

    def add_listener(self, name, listener: PortListener) -> None:
        self._listeners[name] = listener

    def _wait_for_msg(
        self, listener_name: str, wake_up_signal: str
    ) -> Generator[EventExpression, None, None]:
        listener = self._listeners[listener_name]
        if not listener.buffer.has_any():
            yield self.await_signal(sender=listener, signal_label=wake_up_signal)

    def _pop_any_msg(self, listener_name: str) -> Message:
        listener = self._listeners[listener_name]
        return listener.buffer.pop_any()

    def _pop_msg(self, listener_name: str, src_pid: int, dst_pid: int) -> Message:
        listener = self._listeners[listener_name]
        return listener.buffer.pop_msg(src_pid, dst_pid)

    def _has_msg(self, listener_name: str, src_pid: int, dst_pid: int) -> bool:
        listener = self._listeners[listener_name]
        return listener.buffer.has_msg(src_pid, dst_pid)

    def _pop_any_msg_any_source(self, listener_names: List[str]) -> Message:
        for listener_name in listener_names:
            listener = self._listeners[listener_name]
            if listener.buffer.has_any():
                return listener.buffer.pop_any()
        raise RuntimeError

    def _pop_all_messages(self, listener_names: List[str]) -> List[Message]:
        messages = []
        for listener_name in listener_names:
            listener = self._listeners[listener_name]
            if listener.buffer.has_any():
                messages.extend(listener.buffer.pop_all())
        return messages

    def _get_evexpr_for_any_msg(
        self, listener_names: List[str], wake_up_signals: List[str]
    ) -> Optional[EventExpression]:
        # Returns None if there are already messages and no event expression is needed.

        # TODO rewrite two separate lists as function arguments

        # First check if there is any listener with messages in their buffer.
        for listener_name, wake_up_signal in zip(listener_names, wake_up_signals):
            listener = self._listeners[listener_name]
            if listener.buffer.has_any():
                return None

        # Else, get an EventExpression for each listener.
        expressions: List[EventExpression] = []

        for listener_name, wake_up_signal in zip(listener_names, wake_up_signals):
            listener = self._listeners[listener_name]
            assert not listener.buffer.has_any()  # already checked this above
            ev_expr = self.await_signal(sender=listener, signal_label=wake_up_signal)
            expressions.append(ev_expr)

        # Create a union of all expressions.
        assert len(expressions) > 0
        union = expressions[0]
        for i in range(1, len(expressions)):
            union = union | expressions[i]  # type: ignore

        return union

    def _handle_msg_evexpr(
        self, evexpr: EventExpression, listener_names: List[str]
    ) -> Generator[EventExpression, None, None]:
        # Count the number of listeners that has messages in their buffers.
        # This number is equal to the number of events that have fired.
        ev_count = 0
        for listener_name in listener_names:
            listener = self._listeners[listener_name]

            if listener.buffer.has_any():
                ev_count += 1

        # There *must* be at least one event that has fired since we yielded.
        assert ev_count > 0
        # "Flush away" the events that were also in the union.
        # We already yielded on one (above), but need to yield on the rest.
        for _ in range(ev_count - 1):
            yield evexpr

    def _wait_for_msg_any_source(
        self, listener_names: List[str], wake_up_signals: List[str]
    ) -> Generator[EventExpression, None, None]:
        # Get an event expression for any one of the listeners getting a message.
        evexpr = self._get_evexpr_for_any_msg(listener_names, wake_up_signals)

        # If None, there are already messages available.
        if evexpr is None:
            return

        # Wait until at least one of the events in the evexpr happens.
        # (i.e. wait until at least one message arrives)
        yield evexpr

        yield from self._handle_msg_evexpr(evexpr, listener_names)

    def start(self) -> None:
        super().start()
        for listener in self._listeners.values():
            listener.start()

    def stop(self) -> None:
        for listener in self._listeners.values():
            listener.stop()
        super().stop()



from __future__ import annotations

from typing import Generator

import netsquid as ns
from netsquid.components import ClassicalChannel
from netsquid.nodes import Node

from pydynaa import EventExpression
from qoala.lang.ehi import EhiNetworkInfo
from qoala.runtime.message import Message
from qoala.sim.netstack import NetstackComponent, NetstackInterface


class MockNetstackInterface(NetstackInterface):
    def __init__(self, comp: NetstackComponent, ehi_network: EhiNetworkInfo) -> None:
        super().__init__(comp, ehi_network, None, None)


def create_netstackcomp(num_other_nodes: int) -> NetstackComponent:
    node = Node(name="alice", ID=0)

    nodes = {id: f"node_{id}" for id in range(1, num_other_nodes + 1)}
    nodes[0] = "alice"
    ehi_network = EhiNetworkInfo.only_nodes(nodes)

    return NetstackComponent(node, ehi_network)


def test_no_other_nodes():
    comp = create_netstackcomp(num_other_nodes=0)

    # should have host_{in|out}, qnos_{in|out}, qnos_mem_{in|out}, entdist_{in|out} ports
    assert len(comp.ports) == 8
    assert "host_in" in comp.ports
    assert "host_out" in comp.ports
    assert "qnos_in" in comp.ports
    assert "qnos_out" in comp.ports
    assert "qnos_mem_in" in comp.ports
    assert "qnos_mem_out" in comp.ports
    assert "entdist_in" in comp.ports
    assert "entdist_out" in comp.ports


def test_one_other_node():
    comp = create_netstackcomp(num_other_nodes=1)

    # should have 2 host ports + 4 qnos ports + 2 entdist ports + 2 peer ports
    assert len(comp.ports) == 10
    assert "host_in" in comp.ports
    assert "host_out" in comp.ports
    assert "qnos_in" in comp.ports
    assert "qnos_out" in comp.ports
    assert "qnos_mem_in" in comp.ports
    assert "qnos_mem_out" in comp.ports
    assert "entdist_in" in comp.ports
    assert "entdist_out" in comp.ports

    assert "peer_node_1_in" in comp.ports
    assert "peer_node_1_out" in comp.ports

    # Test properties
    assert comp.peer_in_port("node_1") == comp.ports["peer_node_1_in"]
    assert comp.peer_out_port("node_1") == comp.ports["peer_node_1_out"]


def test_many_other_nodes():
    comp = create_netstackcomp(num_other_nodes=5)

    # should have 2 host ports + 4 qnos ports + 2 entdist ports + 5 * 2 peer ports
    assert len(comp.ports) == 18
    assert "host_in" in comp.ports
    assert "host_out" in comp.ports
    assert "qnos_in" in comp.ports
    assert "qnos_out" in comp.ports
    assert "qnos_mem_in" in comp.ports
    assert "qnos_mem_out" in comp.ports
    assert "entdist_in" in comp.ports
    assert "entdist_out" in comp.ports

    for i in range(1, 6):
        assert f"peer_node_{i}_in" in comp.ports
        assert f"peer_node_{i}_out" in comp.ports
        # Test properties
        assert comp.peer_in_port(f"node_{i}") == comp.ports[f"peer_node_{i}_in"]
        assert comp.peer_out_port(f"node_{i}") == comp.ports[f"peer_node_{i}_out"]


def test_connection():
    ns.sim_reset()

    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    ehi_network = EhiNetworkInfo.only_nodes({alice.ID: alice.name, bob.ID: bob.name})

    alice_comp = NetstackComponent(alice, ehi_network)
    bob_comp = NetstackComponent(bob, ehi_network)

    alice_comp.peer_out_port("bob").connect(bob_comp.peer_in_port("alice"))
    alice_comp.peer_in_port("bob").connect(bob_comp.peer_out_port("alice"))

    class AliceNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_peer_msg("bob", Message(0, 0, "hello"))

    class BobNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            msg = yield from self.receive_peer_msg("alice")
            assert msg.content == "hello"

    alice_intf = AliceNetstackInterface(alice_comp, ehi_network)
    bob_intf = BobNetstackInterface(bob_comp, ehi_network)

    alice_intf.start()
    bob_intf.start()

    ns.sim_run()


def test_connection_with_channel():
    ns.sim_reset()

    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    ehi_network = EhiNetworkInfo.only_nodes({alice.ID: alice.name, bob.ID: bob.name})

    alice_comp = NetstackComponent(alice, ehi_network)
    bob_comp = NetstackComponent(bob, ehi_network)

    channel_ab = ClassicalChannel("chan_ab", delay=1000)
    channel_ba = ClassicalChannel("chan_ba", delay=1000)

    alice_comp.peer_out_port("bob").connect(channel_ab.ports["send"])
    channel_ab.ports["recv"].connect(bob_comp.peer_in_port("alice"))

    bob_comp.peer_out_port("alice").connect(channel_ba.ports["send"])
    channel_ba.ports["recv"].connect(alice_comp.peer_in_port("bob"))

    class AliceNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_peer_msg("bob", Message(0, 0, "hello"))

    class BobNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            assert ns.sim_time() == 0
            msg = yield from self.receive_peer_msg("alice")
            assert msg.content == "hello"
            assert ns.sim_time() == 1000

    alice_intf = AliceNetstackInterface(alice_comp, ehi_network)
    bob_intf = BobNetstackInterface(bob_comp, ehi_network)

    alice_intf.start()
    bob_intf.start()

    ns.sim_run()


def test_three_way_connection():
    ns.sim_reset()

    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    charlie = Node(name="charlie", ID=2)
    ehi_network = EhiNetworkInfo.only_nodes(
        {alice.ID: alice.name, bob.ID: bob.name, charlie.ID: charlie.name}
    )

    alice_comp = NetstackComponent(alice, ehi_network)
    bob_comp = NetstackComponent(bob, ehi_network)
    charlie_comp = NetstackComponent(charlie, ehi_network)

    alice_comp.peer_out_port("bob").connect(bob_comp.peer_in_port("alice"))
    alice_comp.peer_in_port("bob").connect(bob_comp.peer_out_port("alice"))
    alice_comp.peer_out_port("charlie").connect(charlie_comp.peer_in_port("alice"))
    alice_comp.peer_in_port("charlie").connect(charlie_comp.peer_out_port("alice"))
    bob_comp.peer_out_port("charlie").connect(charlie_comp.peer_in_port("bob"))
    bob_comp.peer_in_port("charlie").connect(charlie_comp.peer_out_port("bob"))

    class AliceNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_peer_msg("bob", Message(0, 0, "hello bob"))
            self.send_peer_msg("charlie", Message(0, 0, "hello charlie"))

    class BobNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            msg = yield from self.receive_peer_msg("alice")
            assert msg.content == "hello bob"

    class CharlieNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            msg = yield from self.receive_peer_msg("alice")
            assert msg.content == "hello charlie"

    alice_intf = AliceNetstackInterface(alice_comp, ehi_network)
    bob_intf = BobNetstackInterface(bob_comp, ehi_network)
    charlie_intf = CharlieNetstackInterface(charlie_comp, ehi_network)

    alice_intf.start()
    bob_intf.start()
    charlie_intf.start()

    ns.sim_run()


def test_connection_with_pids():
    ns.sim_reset()

    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    ehi_network = EhiNetworkInfo.only_nodes({alice.ID: alice.name, bob.ID: bob.name})

    alice_comp = NetstackComponent(alice, ehi_network)
    bob_comp = NetstackComponent(bob, ehi_network)

    channel_ab = ClassicalChannel("chan_ab", delay=1000)
    channel_ba = ClassicalChannel("chan_ba", delay=1000)

    alice_comp.peer_out_port("bob").connect(channel_ab.ports["send"])
    channel_ab.ports["recv"].connect(bob_comp.peer_in_port("alice"))

    bob_comp.peer_out_port("alice").connect(channel_ba.ports["send"])
    channel_ba.ports["recv"].connect(alice_comp.peer_in_port("bob"))

    class AliceHostInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            # Send message from PID 0 (alice) to PID 0 (bob)
            self.send_peer_msg("bob", Message(0, 0, "hello (0, 0)"))

            yield from self.wait(500)

            # Send message from PID 2 (alice) to PID 1 (bob)
            self.send_peer_msg("bob", Message(2, 1, "hello (2, 1)"))

    class BobHostInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            assert ns.sim_time() == 0
            msg00 = yield from self.receive_peer_msg("alice")
            assert msg00 == Message(0, 0, "hello (0, 0)")
            assert ns.sim_time() == 1000

            msg21 = yield from self.receive_peer_msg("alice")
            print(f"{self.name}: received msg (0, 0) with content: {msg00.content}")
            print(f"{self.name}: received msg (2, 1) with content: {msg21.content}")
            assert msg21 == Message(2, 1, "hello (2, 1)")

    alice_intf = AliceHostInterface(alice_comp, ehi_network)
    bob_intf = BobHostInterface(bob_comp, ehi_network)

    alice_intf.start()
    bob_intf.start()

    ns.sim_run()


if __name__ == "__main__":
    test_no_other_nodes()
    test_one_other_node()
    test_many_other_nodes()

    test_connection()
    test_connection_with_channel()
    test_three_way_connection()
    test_connection_with_pids()



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

import netsquid as ns
from netqasm.lang.parsing import parse_text_subroutine

from qoala.lang.ehi import UnitModule
from qoala.lang.parse import LocalRoutineParser
from qoala.lang.program import LocalRoutine, ProgramMeta, QoalaProgram
from qoala.lang.routine import RoutineMetadata
from qoala.runtime.lhi import LhiTopology, LhiTopologyBuilder
from qoala.runtime.lhi_to_ehi import LhiConverter
from qoala.runtime.memory import ProgramMemory
from qoala.runtime.message import Message
from qoala.runtime.ntf import NvNtf
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.runtime.sharedmem import MemAddr
from qoala.sim.memmgr import MemoryManager
from qoala.sim.process import QoalaProcess
from qoala.sim.qdevice import QDevice
from qoala.sim.qnos import GenericProcessor, QnosInterface, QnosLatencies, QnosProcessor
from qoala.util.tests import netsquid_run, yield_from

MOCK_QNOS_RET_REG = "R0"
MOCK_QNOS_RET_VALUE = 7


@dataclass(frozen=True)
class InterfaceEvent:
    peer: str
    msg: Message


@dataclass(frozen=True)
class FlushEvent:
    pass


@dataclass(frozen=True)
class SignalEvent:
    pass


class MockQDevice(QDevice):
    def __init__(self, topology: LhiTopology) -> None:
        self._topology = topology

    def set_mem_pos_in_use(self, id: int, in_use: bool) -> None:
        pass


@dataclass
class MockNetstackResultInfo:
    pid: int
    array_id: int
    start_idx: int
    end_idx: int


class MockQnosInterface(QnosInterface):
    def __init__(
        self,
        qdevice: QDevice,
    ) -> None:
        self.send_events: List[InterfaceEvent] = []
        self.recv_events: List[InterfaceEvent] = []
        self.flush_events: List[FlushEvent] = []
        self.signal_events: List[SignalEvent] = []

        self._qdevice = qdevice
        self._memmgr = MemoryManager("alice", self._qdevice)

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        self.send_events.append(InterfaceEvent(peer, msg))

    def signal_memory_freed(self) -> None:
        self.signal_events.append(SignalEvent())

    @property
    def name(self) -> str:
        return "mock"


def create_program(
    subroutines: Optional[Dict[str, LocalRoutine]] = None,
    meta: Optional[ProgramMeta] = None,
) -> QoalaProgram:
    if subroutines is None:
        subroutines = {}
    if meta is None:
        meta = ProgramMeta.empty("prog")
    return QoalaProgram(blocks=[], local_routines=subroutines, meta=meta)


def create_process(
    pid: int,
    program: QoalaProgram,
    unit_module: UnitModule,
    inputs: Optional[ProgramInput] = None,
) -> QoalaProcess:
    if inputs is None:
        inputs = ProgramInput({})

    instance = ProgramInstance(
        pid=pid,
        program=program,
        inputs=inputs,
        unit_module=unit_module,
    )
    mem = ProgramMemory(pid=pid)

    process = QoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={},
        epr_sockets=program.meta.epr_sockets,
        result=ProgramResult(values={}),
    )
    return process


def create_process_with_subrt(
    pid: int,
    subrt_text: str,
    unit_module: UnitModule,
    inputs: Optional[ProgramInput] = None,
) -> QoalaProcess:
    subrt = parse_text_subroutine(subrt_text)
    metadata = RoutineMetadata.use_none()
    iqoala_subrt = LocalRoutine("subrt", subrt, return_vars=[], metadata=metadata)
    meta = ProgramMeta.empty("alice")
    meta.epr_sockets = {0: "bob"}
    program = create_program(subroutines={"subrt": iqoala_subrt}, meta=meta)
    return create_process(pid, program, unit_module, inputs)


def create_process_with_local_routine(
    pid: int,
    routine_text: str,
    unit_module: UnitModule,
    inputs: Optional[ProgramInput] = None,
) -> QoalaProcess:
    routines = LocalRoutineParser(routine_text).parse()
    meta = ProgramMeta.empty("alice")
    meta.epr_sockets = {0: "bob"}
    program = create_program(subroutines=routines, meta=meta)
    return create_process(pid, program, unit_module, inputs)


def execute_process(
    processor: GenericProcessor,
    process: QoalaProcess,
    input_addr: Optional[MemAddr] = None,
    result_addr: Optional[MemAddr] = None,
) -> int:
    if input_addr is None:
        input_addr = MemAddr(0)
    if result_addr is None:
        result_addr = MemAddr(0)

    all_routines = process.program.local_routines
    routine = all_routines["subrt"]
    inputs = process.inputs.values
    processor.instantiate_routine(process, routine, inputs, input_addr, result_addr)

    netqasm_instructions = routine.subroutine.instructions

    instr_count = 0

    instr_idx = 0
    while instr_idx < len(netqasm_instructions):
        instr_count += 1
        instr_idx = yield_from(
            processor.assign_routine_instr(process, "subrt", instr_idx)
        )
    return instr_count


def execute_process_with_latencies(
    processor: GenericProcessor,
    process: QoalaProcess,
    input_addr: Optional[MemAddr] = None,
    result_addr: Optional[MemAddr] = None,
) -> int:
    if input_addr is None:
        input_addr = MemAddr(0)
    if result_addr is None:
        result_addr = MemAddr(0)
    all_routines = process.program.local_routines
    routine = all_routines["subrt"]

    inputs = process.inputs.values
    processor.instantiate_routine(process, routine, inputs, input_addr, result_addr)

    netqasm_instructions = routine.subroutine.instructions

    instr_count = 0

    instr_idx = 0
    while instr_idx < len(netqasm_instructions):
        instr_count += 1
        instr_idx = netsquid_run(
            processor.assign_routine_instr(process, "subrt", instr_idx)
        )
    return instr_count


def execute_multiple_processes(
    processor: GenericProcessor, processes: List[QoalaProcess]
) -> None:
    for proc in processes:
        all_routines = proc.program.local_routines
        routine = all_routines["subrt"]
        # input/result arrays not used
        # TODO: add tests that do use these
        inputs = proc.inputs.values
        processor.instantiate_routine(proc, routine, inputs, MemAddr(0), MemAddr(0))
        netqasm_instructions = routine.subroutine.instructions
        for i in range(len(netqasm_instructions)):
            yield_from(processor.assign_routine_instr(proc, "subrt", i))


def setup_components(
    topology: LhiTopology, latencies: QnosLatencies = QnosLatencies.all_zero()
) -> Tuple[QnosProcessor, UnitModule]:
    qdevice = MockQDevice(topology)
    ehi = LhiConverter.to_ehi(topology, ntf=NvNtf())
    unit_module = UnitModule.from_full_ehi(ehi)
    interface = MockQnosInterface(qdevice)
    processor = QnosProcessor(interface, latencies)
    return processor, unit_module


def uniform_topology(num_qubits: int) -> LhiTopology:
    return LhiTopologyBuilder.perfect_uniform(num_qubits, [], 0, [], 0)


def star_topology(num_qubits: int) -> LhiTopology:
    return LhiTopologyBuilder.perfect_star(num_qubits, [], 0, [], 0, [], 0)


def native_instr_count(subrt_text: str) -> int:
    # count the number of instructions in the subroutine when the subrt text
    # is parsed and compiled (which may lead to additional instructions)
    parsed_subrt = parse_text_subroutine(subrt_text)
    return len(parsed_subrt.instructions)


def test_set_reg():
    processor, unit_module = setup_components(star_topology(2))

    subrt = """
    set R0 17
    """
    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)
    assert process.prog_memory.qnos_mem.get_reg_value("R0") == 17


def test_set_reg_with_latencies():
    ns.sim_reset()

    processor, unit_module = setup_components(
        star_topology(2), latencies=QnosLatencies(qnos_instr_time=5e3)
    )

    subrt = """
    set R0 17
    """
    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)

    assert ns.sim_time() == 0
    execute_process_with_latencies(processor, process)
    assert ns.sim_time() == 5e3

    assert process.prog_memory.qnos_mem.get_reg_value("R0") == 17


def test_add():
    processor, unit_module = setup_components(star_topology(2))

    subrt = """
    set R0 2
    set R1 5
    add R2 R0 R1
    """
    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)
    assert process.prog_memory.qnos_mem.get_reg_value("R2") == 7


def test_add_with_latencies():
    ns.sim_reset()

    processor, unit_module = setup_components(
        star_topology(2), latencies=QnosLatencies(qnos_instr_time=5e3)
    )

    subrt = """
    set R0 2
    set R1 5
    add R2 R0 R1
    """
    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    assert native_instr_count(subrt) == 3

    assert ns.sim_time() == 0
    execute_process_with_latencies(processor, process)
    assert ns.sim_time() == 5e3 * 3

    assert process.prog_memory.qnos_mem.get_reg_value("R2") == 7


def test_mul():
    processor, unit_module = setup_components(star_topology(2))

    subrt = """
    set R0 2
    set R1 5
    mul R2 R0 R1
    """
    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)
    assert process.prog_memory.qnos_mem.get_reg_value("R2") == 10


def test_mul_with_latencies():
    ns.sim_reset()

    processor, unit_module = setup_components(
        star_topology(2), latencies=QnosLatencies(qnos_instr_time=5e3)
    )

    subrt = """
    set R0 2
    set R1 5
    mul R2 R0 R1
    """
    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)

    assert ns.sim_time() == 0
    execute_process_with_latencies(processor, process)
    assert ns.sim_time() == 5e3 * 3

    assert process.prog_memory.qnos_mem.get_reg_value("R2") == 10


def test_div():
    processor, unit_module = setup_components(star_topology(2))

    subrt = """
    set R0 15
    set R1 3
    div R2 R0 R1
    """
    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)
    assert process.prog_memory.qnos_mem.get_reg_value("R2") == 5


def test_div_with_latencies():
    ns.sim_reset()

    processor, unit_module = setup_components(
        star_topology(2), latencies=QnosLatencies(qnos_instr_time=5e3)
    )

    subrt = """
    set R0 15
    set R1 3
    div R2 R0 R1
    """
    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)

    assert ns.sim_time() == 0
    execute_process_with_latencies(processor, process)
    assert ns.sim_time() == 5e3 * 3

    assert process.prog_memory.qnos_mem.get_reg_value("R2") == 5


def test_div_rounded():
    processor, unit_module = setup_components(star_topology(2))

    subrt = """
    set R0 16
    set R1 3
    div R2 R0 R1
    """
    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)
    assert process.prog_memory.qnos_mem.get_reg_value("R2") == 5


def test_rem():
    processor, unit_module = setup_components(star_topology(2))

    subrt = """
    set R0 16
    set R1 3
    rem R2 R0 R1
    """
    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)
    assert process.prog_memory.qnos_mem.get_reg_value("R2") == 1


def test_no_branch():
    processor, unit_module = setup_components(star_topology(2))

    subrt = """
    set R3 3
    set R0 0
    beq R3 R0 LABEL1
    set R1 1
    add C0 R3 R1
LABEL1:
    """
    assert native_instr_count(subrt) == 5

    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    instr_count = execute_process(processor, process)

    assert instr_count == 5
    assert process.prog_memory.qnos_mem.get_reg_value("C0") == 4


def test_branch():
    processor, unit_module = setup_components(star_topology(2))

    subrt = """
    set R3 3
    set C3 3
    beq R3 C3 LABEL1
    set R1 1
    add C0 R3 R1
LABEL1:
    """
    assert native_instr_count(subrt) == 5

    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)
    instr_count = execute_process(processor, process)

    assert instr_count == 3
    assert process.prog_memory.qnos_mem.get_reg_value("C0") == 0


def test_branch_with_latencies():
    ns.sim_reset()

    processor, unit_module = setup_components(
        star_topology(2), latencies=QnosLatencies(qnos_instr_time=5e3)
    )

    subrt = """
    set R3 3
    set C3 3
    beq R3 C3 LABEL1
    set R1 1
    add C0 R3 R1
LABEL1:
    """
    assert native_instr_count(subrt) == 5

    process = create_process_with_subrt(0, subrt, unit_module)
    processor._interface.memmgr.add_process(process)

    assert ns.sim_time() == 0
    instr_count = execute_process_with_latencies(processor, process)
    assert instr_count == 3
    assert ns.sim_time() == 5e3 * 3

    assert process.prog_memory.qnos_mem.get_reg_value("C0") == 0


def test_program_inputs():
    processor, unit_module = setup_components(star_topology(2))

    subrt = """
    set R0 {global_arg}
    """
    inputs = ProgramInput({"global_arg": 3})

    process = create_process_with_subrt(0, subrt, unit_module, inputs)
    processor._interface.memmgr.add_process(process)
    execute_process(processor, process)

    assert process.qnos_mem.get_reg_value("R0") == 3


def test_program_inputs_with_latencies():
    ns.sim_reset()

    processor, unit_module = setup_components(
        star_topology(2), latencies=QnosLatencies(qnos_instr_time=5e3)
    )

    subrt = """
    set R0 {global_arg}
    """
    inputs = ProgramInput({"global_arg": 3})

    process = create_process_with_subrt(0, subrt, unit_module, inputs)
    processor._interface.memmgr.add_process(process)

    assert ns.sim_time() == 0
    execute_process_with_latencies(processor, process)
    assert ns.sim_time() == 5e3

    assert process.qnos_mem.get_reg_value("R0") == 3


def test_program_routine_params():
    processor, unit_module = setup_components(star_topology(2))

    routine = """
SUBROUTINE subrt
    params: 
    returns: 
    uses: 
    keeps:
    request: 
  NETQASM_START
    load R0 @input[0]
  NETQASM_END
    """

    process = create_process_with_local_routine(0, routine, unit_module)
    processor._interface.memmgr.add_process(process)

    shared_mem = process.prog_memory.shared_mem
    input_addr = shared_mem.allocate_lr_in(1)
    shared_mem.write_lr_in(input_addr, [3])

    execute_process(processor, process, input_addr=input_addr, result_addr=0)

    assert process.qnos_mem.get_reg_value("R0") == 3


def test_program_routine_params_with_latencies():
    ns.sim_reset()

    processor, unit_module = setup_components(
        star_topology(2), latencies=QnosLatencies(qnos_instr_time=5e3)
    )

    routine = """
SUBROUTINE subrt
    params: 
    returns: 
    uses: 
    keeps:
    request: 
  NETQASM_START
    load R0 @input[0]
  NETQASM_END
    """

    process = create_process_with_local_routine(0, routine, unit_module)
    processor._interface.memmgr.add_process(process)

    shared_mem = process.prog_memory.shared_mem
    input_addr = shared_mem.allocate_lr_in(1)
    shared_mem.write_lr_in(input_addr, [3])

    assert ns.sim_time() == 0
    execute_process_with_latencies(
        processor, process, input_addr=input_addr, result_addr=None
    )
    # The load R0 @input[0] is converted to 2 instructions as follows:
    #  set R1 0
    #  load R0 @input[R1]
    assert ns.sim_time() == 5e3 * 2

    assert process.qnos_mem.get_reg_value("R0") == 3


def test_program_routine_params_and_results():
    processor, unit_module = setup_components(star_topology(2))

    # TODO: fill in params and returns when Host/Array conversion is implemented
    routine = """
SUBROUTINE subrt
    params:
    returns: 
    uses: 
    keeps:
    request: 
  NETQASM_START
    load R0 @input[0]
    load R1 @input[1]
    add C0 R0 R0
    add C1 R1 R1
    store C0 @output[0]
    store C1 @output[1]
  NETQASM_END
    """

    process = create_process_with_local_routine(0, routine, unit_module)
    processor._interface.memmgr.add_process(process)

    shared_mem = process.prog_memory.shared_mem

    input_addr = shared_mem.allocate_lr_in(2)
    shared_mem.write_lr_in(input_addr, [3, 7])

    result_addr = shared_mem.allocate_lr_out(2)

    execute_process(processor, process, input_addr=input_addr, result_addr=result_addr)

    assert process.qnos_mem.get_reg_value("C0") == 6
    assert process.qnos_mem.get_reg_value("C1") == 14


def test_program_routine_params_and_results_with_latecies():
    ns.sim_reset()

    processor, unit_module = setup_components(
        star_topology(2), latencies=QnosLatencies(qnos_instr_time=5e3)
    )
    # TODO: fill in params and returns when Host/Array conversion is implemented
    routine = """
SUBROUTINE subrt
    params:
    returns: 
    uses: 
    keeps:
    request: 
  NETQASM_START
    load R0 @input[0]
    load R1 @input[1]
    add C0 R0 R0
    add C1 R1 R1
    store C0 @output[0]
    store C1 @output[1]
  NETQASM_END
    """

    process = create_process_with_local_routine(0, routine, unit_module)
    processor._interface.memmgr.add_process(process)

    shared_mem = process.prog_memory.shared_mem

    input_addr = shared_mem.allocate_lr_in(2)
    shared_mem.write_lr_in(input_addr, [3, 7])

    result_addr = shared_mem.allocate_lr_out(2)

    assert ns.sim_time() == 0
    execute_process_with_latencies(
        processor, process, input_addr=input_addr, result_addr=result_addr
    )

    # load and store instructions are converted to 2 instructions each
    assert ns.sim_time() == 5e3 * 10

    assert process.qnos_mem.get_reg_value("C0") == 6
    assert process.qnos_mem.get_reg_value("C1") == 14


if __name__ == "__main__":
    test_set_reg()
    test_set_reg_with_latencies()
    test_add()
    test_add_with_latencies()
    test_mul()
    test_mul_with_latencies()
    test_div()
    test_div_with_latencies()
    test_div_rounded()
    test_rem()
    test_no_branch()
    test_branch()
    test_branch_with_latencies()
    test_program_inputs()
    test_program_inputs_with_latencies()
    test_program_routine_params()
    test_program_routine_params_with_latencies()
    test_program_routine_params_and_results()
    test_program_routine_params_and_results_with_latecies()



import os
import subprocess


def main():
    path_to_here = os.path.dirname(os.path.abspath(__file__))
    for root, _, files in os.walk(path_to_here):
        for filename in files:
            if filename.startswith("example") and filename.endswith(".py"):
                filepath = os.path.join(root, filename)
                print(f"Running example {filepath}")
                result = subprocess.run(
                    ["python3", filepath],
                    stdout=subprocess.DEVNULL,
                )
                if result.returncode != 0:
                    raise RuntimeError(f"Example {filepath} failed!")

    print("All examples work!")


if __name__ == "__main__":
    main()



from typing import Generator

import netsquid as ns
from netsquid.components.cchannel import ClassicalChannel

from pydynaa import EventExpression
from qoala.lang.ehi import EhiNetworkInfo
from qoala.runtime.lhi import LhiLatencies, LhiTopologyBuilder
from qoala.runtime.message import Message
from qoala.runtime.ntf import GenericNtf
from qoala.sim.procnode import ProcNode
from qoala.sim.procnodecomp import ProcNodeComponent


def create_procnodecomp(num_other_nodes: int) -> ProcNodeComponent:
    nodes = {id: f"node_{id}" for id in range(1, num_other_nodes + 1)}
    nodes[0] = "alice"
    ehi_network = EhiNetworkInfo.only_nodes(nodes)

    return ProcNodeComponent(name="alice", qprocessor=None, ehi_network=ehi_network)


def test_no_other_nodes():
    comp = create_procnodecomp(num_other_nodes=0)

    # should not have 2 entdist ports
    assert len(comp.ports) == 2
    assert "entdist_in" in comp.ports
    assert "entdist_out" in comp.ports

    assert len(comp.subcomponents) == 3


def test_one_other_node():
    comp = create_procnodecomp(num_other_nodes=1)

    # should have 2 host peer ports + 2 netstack peer ports + 2 entdist ports
    assert len(comp.ports) == 6
    assert "host_peer_node_1_in" in comp.ports
    assert "host_peer_node_1_out" in comp.ports
    assert "netstack_peer_node_1_in" in comp.ports
    assert "netstack_peer_node_1_out" in comp.ports
    assert "entdist_in" in comp.ports
    assert "entdist_out" in comp.ports

    # Test properties
    assert comp.host_peer_in_port("node_1") == comp.ports["host_peer_node_1_in"]
    assert comp.host_peer_out_port("node_1") == comp.ports["host_peer_node_1_out"]
    assert comp.netstack_peer_in_port("node_1") == comp.ports["netstack_peer_node_1_in"]
    assert (
        comp.netstack_peer_out_port("node_1") == comp.ports["netstack_peer_node_1_out"]
    )

    assert len(comp.subcomponents) == 3


def test_many_other_nodes():
    comp = create_procnodecomp(num_other_nodes=5)

    # should 5 * 4 peer ports + 2 entdist ports
    assert len(comp.ports) == 22

    for i in range(1, 6):
        assert f"host_peer_node_{i}_in" in comp.ports
        assert f"host_peer_node_{i}_out" in comp.ports
        assert f"netstack_peer_node_{i}_in" in comp.ports
        assert f"netstack_peer_node_{i}_out" in comp.ports
        # Test properties
        assert (
            comp.host_peer_in_port(f"node_{i}") == comp.ports[f"host_peer_node_{i}_in"]
        )
        assert (
            comp.host_peer_out_port(f"node_{i}")
            == comp.ports[f"host_peer_node_{i}_out"]
        )
        assert (
            comp.netstack_peer_in_port(f"node_{i}")
            == comp.ports[f"netstack_peer_node_{i}_in"]
        )
        assert (
            comp.netstack_peer_out_port(f"node_{i}")
            == comp.ports[f"netstack_peer_node_{i}_out"]
        )

    assert len(comp.subcomponents) == 3


def test_connection_with_channel():
    ns.sim_reset()

    nodes = {0: "alice", 1: "bob"}

    ehi_network = EhiNetworkInfo.only_nodes(nodes)
    alice = ProcNodeComponent("alice", None, ehi_network=ehi_network, node_id=0)
    bob = ProcNodeComponent("bob", None, ehi_network=ehi_network, node_id=1)

    channel_ab = ClassicalChannel("chan_ab", delay=1000)

    alice.host_peer_out_port("bob").connect(channel_ab.ports["send"])
    channel_ab.ports["recv"].connect(bob.host_peer_in_port("alice"))
    alice.host_peer_in_port("bob").connect(bob.host_peer_out_port("alice"))

    class AliceProcnode(ProcNode):
        def run(self) -> Generator[EventExpression, None, None]:
            self.host.interface.send_peer_msg("bob", Message(0, 0, "hello"))

    class BobProcnode(ProcNode):
        def run(self) -> Generator[EventExpression, None, None]:
            assert ns.sim_time() == 0
            msg = yield from self.host.interface.receive_peer_msg("alice")
            assert ns.sim_time() == 1000
            assert msg == Message(0, 0, "hello")

    topology = LhiTopologyBuilder.perfect_uniform_default_gates(1)
    latencies = LhiLatencies.all_zero()
    ntf = GenericNtf()
    alice_proc = AliceProcnode(
        "alice", None, topology, latencies, ntf, ehi_network, node=alice
    )
    bob_proc = BobProcnode("bob", None, topology, latencies, ntf, ehi_network, node=bob)

    alice_proc.start()
    bob_proc.start()

    ns.sim_run()


if __name__ == "__main__":
    test_no_other_nodes()
    test_one_other_node()
    test_many_other_nodes()

    test_connection_with_channel()



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Generator, List, Optional, Tuple

from pydynaa import EventExpression
from qoala.lang.ehi import EhiNetworkInfo
from qoala.runtime.message import Message
from qoala.sim.componentprot import ComponentProtocol, PortListener
from qoala.sim.events import (
    EVENT_WAIT,
    SIGNAL_HOST_HOST_MSG,
    SIGNAL_NSTK_HOST_MSG,
    SIGNAL_QNOS_HOST_MSG,
)
from qoala.sim.host.hostcomp import HostComponent


@dataclass
class HostLatencies:
    host_instr_time: float = 0  # duration of classical Host instr execution
    host_peer_latency: float = 0  # processing time for messages from remote node

    @classmethod
    def all_zero(cls) -> HostLatencies:
        # NOTE: can also just use HostLatencies() which will default all values to 0
        # However, using this classmethod makes this behavior more explicit and clear.
        return HostLatencies(0, 0)


class HostInterface(ComponentProtocol):
    """NetSquid protocol representing a Host."""

    def __init__(
        self,
        comp: HostComponent,
        ehi_network: EhiNetworkInfo,
    ) -> None:
        """Host protocol constructor.

        :param comp: NetSquid component representing the Host
        """
        name = f"{comp.name}_protocol"
        super().__init__(name=name, comp=comp)
        self._comp = comp

        self._ehi_network = ehi_network

        self.add_listener(
            "qnos",
            PortListener(self._comp.qnos_in_port, SIGNAL_QNOS_HOST_MSG),
        )
        self.add_listener(
            "netstack",
            PortListener(self._comp.netstack_in_port, SIGNAL_NSTK_HOST_MSG),
        )
        self._listener_names = []
        self._signal_names = []
        for peer in self._ehi_network.nodes.values():
            if peer == self._comp.node_name:
                continue
            self.add_listener(
                f"peer_{peer}",
                PortListener(
                    self._comp.peer_in_port(peer), f"{SIGNAL_HOST_HOST_MSG}_{peer}"
                ),
            )
            self._listener_names.append(f"peer_{peer}")
            self._signal_names.append(f"{SIGNAL_HOST_HOST_MSG}_{peer}")

        # If the last instruction executed for the program instance was a jump,
        # this will be the index of the block(in program's list of blocks) to jump to, otherwise None.
        self._program_instance_jumps: Dict[int, int] = {}  # pid => block index

    @property
    def program_instance_jumps(self) -> Dict[int, int]:
        return self._program_instance_jumps

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        self._logger.info(f"sending message {msg} to {peer}")
        self._comp.peer_out_port(peer).tx_output(msg)

    def get_available_messages(self, peer: str) -> List[Tuple[int, int]]:
        listener = self._listeners[f"peer_{peer}"]
        return listener.buffer.get_all()

    def wait_for_msg(self, peer: str) -> Generator[EventExpression, None, None]:
        yield from self._wait_for_msg(f"peer_{peer}", f"{SIGNAL_HOST_HOST_MSG}_{peer}")

    def wait_for_any_msg(self) -> Generator[EventExpression, None, None]:
        yield from self._wait_for_msg_any_source(
            self._listener_names, self._signal_names
        )

    def get_evexpr_for_any_msg(self) -> Optional[EventExpression]:
        return self._get_evexpr_for_any_msg(self._listener_names, self._signal_names)

    def handle_msg_evexpr(
        self, evexpr: EventExpression
    ) -> Generator[EventExpression, None, None]:
        yield from self._handle_msg_evexpr(evexpr, self._listener_names)

    def pop_msg(self, peer: str, src_pid: int, dst_pid: int) -> Message:
        return self._pop_msg(f"peer_{peer}", src_pid, dst_pid)

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        yield from self._wait_for_msg(f"peer_{peer}", f"{SIGNAL_HOST_HOST_MSG}_{peer}")
        return self._pop_any_msg(f"peer_{peer}")

    def wait(self, delta_time: float) -> Generator[EventExpression, None, None]:
        self._schedule_after(delta_time, EVENT_WAIT)
        event_expr = EventExpression(source=self, event_type=EVENT_WAIT)
        yield event_expr



from __future__ import annotations

import datetime
import json
import math
import os
import time
from argparse import ArgumentParser
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import netsquid as ns

from qoala.lang.ehi import UnitModule
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    ClassicalConnectionConfig,
    LatenciesConfig,
    NetworkScheduleConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchInfo, BatchResult, ProgramBatch, ProgramInput
from qoala.sim.build import build_network_from_config
from qoala.sim.network import ProcNodeNetwork


def relative_to_cwd(file: str) -> str:
    return os.path.join(os.path.dirname(__file__), file)


def topology_config(num_qubits: int, t1: int, t2: int) -> TopologyConfig:
    return TopologyConfig.uniform_t1t2_qubits_perfect_gates_default_params(
        num_qubits, t1=t1, t2=t2
    )


def get_client_config(id: int) -> ProcNodeConfig:
    # client only needs 1 qubit
    return ProcNodeConfig(
        node_name=f"client_{id}",
        node_id=id,
        topology=topology_config(1, t1=0, t2=0),
        latencies=LatenciesConfig(
            host_instr_time=500, host_peer_latency=30_000, qnos_instr_time=1000
        ),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )


def get_server_config(
    id: int, num_qubits: int, t1: int, t2: int, sched_latency: float
) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name="server",
        node_id=id,
        topology=topology_config(num_qubits, t1=t1, t2=t2),
        latencies=LatenciesConfig(
            host_instr_time=500,
            host_peer_latency=30_000,
            qnos_instr_time=1000,
            internal_sched_latency=sched_latency,
        ),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )


def create_network(
    server_cfg: ProcNodeConfig,
    client_configs: List[ProcNodeConfig],
    num_clients: int,
    cc: float,
    bin_length: float,
) -> ProcNodeNetwork:
    assert len(client_configs) == num_clients

    node_cfgs = [server_cfg] + client_configs
    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=node_cfgs, link_duration=1000
    )

    pattern = [
        (client_cfg.node_id, 0, server_cfg.node_id, i)
        for i, client_cfg in enumerate(client_configs)
    ]
    network_cfg.netschedule = NetworkScheduleConfig(
        bin_length=bin_length,
        first_bin=0,
        bin_pattern=pattern,
        repeat_period=bin_length * num_clients,
    )

    cconns = [
        ClassicalConnectionConfig.from_nodes(server_cfg.node_id, client_cfg.node_id, cc)
        for client_cfg in client_configs
    ]
    network_cfg.cconns = cconns
    return build_network_from_config(network_cfg)


@dataclass
class BqcResult:
    client_batches: List[Dict[int, ProgramBatch]]
    client_results: List[Dict[int, BatchResult]]


def load_server_program(remote_name: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), "vbqc_server.iqoala")
    with open(path) as file:
        server_text = file.read()
    program = QoalaParser(server_text).parse()

    # Replace "client" by e.g. "client_1"
    program.meta.csockets[0] = remote_name
    program.meta.epr_sockets[0] = remote_name

    return program


def load_client_program() -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), "vbqc_client.iqoala")
    with open(path) as file:
        client_text = file.read()
    return QoalaParser(client_text).parse()


def create_server_batch(
    client_id: int,
    inputs: List[ProgramInput],
    unit_module: UnitModule,
    num_iterations: int,
) -> BatchInfo:
    server_program = load_server_program(remote_name=f"client_{client_id}")
    return BatchInfo(
        program=server_program,
        inputs=inputs,
        unit_module=unit_module,
        num_iterations=num_iterations,
        deadline=0,
    )


def create_client_batch(
    inputs: List[ProgramInput],
    unit_module: UnitModule,
    num_iterations: int,
) -> BatchInfo:
    client_program = load_client_program()
    return BatchInfo(
        program=client_program,
        inputs=inputs,
        unit_module=unit_module,
        num_iterations=num_iterations,
        deadline=0,
    )


def run_bqc(
    alpha: int,
    beta: int,
    theta1: int,
    theta2: int,
    dummy0: int,
    dummy1: int,
    num_iterations: List[int],
    num_clients: int,
    t1: int,
    t2: int,
    cc: float,
    sched_latency: float,
    bin_length: float,
):
    ns.sim_reset()

    # server needs to have 2 qubits per client
    server_num_qubits = num_clients * 2
    server_config = get_server_config(
        id=0, num_qubits=server_num_qubits, t1=t1, t2=t2, sched_latency=sched_latency
    )
    client_configs = [get_client_config(i) for i in range(1, num_clients + 1)]

    network = create_network(
        server_config,
        client_configs,
        num_clients,
        cc=cc,
        bin_length=bin_length,
    )
    server_procnode = network.nodes["server"]
    server_batches: Dict[int, ProgramBatch] = {}  # client ID -> server batch
    client_batches: Dict[int, ProgramBatch] = {}  # client ID -> client batch

    for client_id in range(1, num_clients + 1):
        # index in num_iterations list
        index = client_id - 1

        server_inputs = [
            ProgramInput({"client_id": client_id}) for _ in range(num_iterations[index])
        ]

        server_unit_module = UnitModule.from_full_ehi(server_procnode.memmgr.get_ehi())
        server_batch_info = create_server_batch(
            client_id=client_id,
            inputs=server_inputs,
            unit_module=server_unit_module,
            num_iterations=num_iterations[index],
        )

        server_batches[client_id] = server_procnode.submit_batch(server_batch_info)

    for client_id in range(1, num_clients + 1):
        # index in num_iterations list
        index = client_id - 1

        client_inputs = [
            ProgramInput(
                {
                    "server_id": 0,
                    "alpha": alpha,
                    "beta": beta,
                    "theta1": theta1,
                    "theta2": theta2,
                    "dummy0": dummy0,
                    "dummy1": dummy1,
                }
            )
            for _ in range(num_iterations[index])
        ]

        client_procnode = network.nodes[f"client_{client_id}"]

        client_unit_module = UnitModule.from_full_ehi(client_procnode.memmgr.get_ehi())
        client_batch_info = create_client_batch(
            client_inputs, client_unit_module, num_iterations[index]
        )

        client_batches[client_id] = client_procnode.submit_batch(client_batch_info)
        batch_id = client_batches[client_id].batch_id
        server_pids = [inst.pid for inst in server_batches[client_id].instances]
        # print(
        #     f"client ID: {client_id}, batch ID: {batch_id}, server PIDs: {server_pids}"
        # )
        client_procnode.initialize_processes(
            remote_pids={batch_id: server_pids}, linear=True
        )

    client_pids = {
        server_batches[client_id].batch_id: [
            inst.pid for inst in client_batches[client_id].instances
        ]
        for client_id in range(1, num_clients + 1)
    }
    # print(f"client PIDs: {client_pids}")
    server_procnode.initialize_processes(remote_pids=client_pids, linear=False)

    network.start()
    start_time = ns.sim_time()
    ns.sim_run()
    end_time = ns.sim_time()
    makespan = end_time - start_time

    client_procnodes = [network.nodes[f"client_{i}"] for i in range(1, num_clients + 1)]
    client_batches = [node.get_batches() for node in client_procnodes]

    client_results: List[Dict[int, BatchResult]]
    client_results = [node.scheduler.get_batch_results() for node in client_procnodes]

    return BqcResult(client_batches, client_results), makespan


def check_trap(
    alpha: int,
    beta: int,
    theta1: int,
    theta2: int,
    dummy0: int,
    dummy1: int,
    num_iterations: List[int],
    num_clients: int,
    t1: int,
    t2: int,
    cc: float,
    sched_latency: float,
    bin_length: float,
):
    ns.sim_reset()
    bqc_result, makespan = run_bqc(
        alpha=alpha,
        beta=beta,
        theta1=theta1,
        theta2=theta2,
        dummy0=dummy0,
        dummy1=dummy1,
        num_iterations=num_iterations,
        num_clients=num_clients,
        t1=t1,
        t2=t2,
        cc=cc,
        sched_latency=sched_latency,
        bin_length=bin_length,
    )

    batch_success_probabilities: List[float] = []

    for i in range(num_clients):
        assert len(bqc_result.client_results[i]) == 1
        batch_result = bqc_result.client_results[i][0]
        assert len(bqc_result.client_batches[i]) == 1
        program_batch = bqc_result.client_batches[i][0]
        batch_iterations = program_batch.info.num_iterations

        p1s = [result.values["p1"] for result in batch_result.results]
        p2s = [result.values["p2"] for result in batch_result.results]
        m1s = [result.values["m1"] for result in batch_result.results]
        m2s = [result.values["m2"] for result in batch_result.results]

        if dummy0 == 0:
            # corresponds to "dummy = 1"
            # do normal rotations on qubit 0
            # no rotations on qubit 1
            num_fails = len([(p, m) for (p, m) in zip(p1s, m2s) if p != m])
        else:  # dummy0 = 1
            # corresponds to "dummy = 2"
            # no rotations on qubit 0
            # do normal rotations on qubit 1
            num_fails = len([(p, m) for (p, m) in zip(p2s, m1s) if p != m])

        frac_fail = round(num_fails / batch_iterations, 2)
        batch_success_probabilities.append(1 - frac_fail)

    return batch_success_probabilities, makespan


def compute_succ_prob_trap(
    num_clients: int,
    num_iterations: List[int],
    t1: int,
    t2: int,
    cc: float,
    sched_latency: float,
    bin_length: float,
):
    ns.set_qstate_formalism(ns.qubits.qformalism.QFormalism.DM)

    return check_trap(
        alpha=8,
        beta=24,
        theta1=2,
        theta2=22,
        dummy0=0,
        dummy1=1,
        num_iterations=num_iterations,
        num_clients=num_clients,
        t1=t1,
        t2=t2,
        cc=cc,
        sched_latency=sched_latency,
        bin_length=bin_length,
    )


def bqc_trap(
    num_clients: int,
    num_iterations: int,
    t1: int,
    t2: int,
    cc: float,
    sched_latency: float,
    bin_length: float,
) -> Tuple[float, float]:
    succ_probs, makespan = compute_succ_prob_trap(
        num_clients=num_clients,
        num_iterations=[num_iterations] * num_clients,
        t1=t1,
        t2=t2,
        cc=cc,
        sched_latency=sched_latency,
        bin_length=bin_length,
    )
    # print(f"success probabilities: {succ_probs}")
    # print(f"makespan: {makespan}")

    avg_succ_prob = sum(succ_probs) / len(succ_probs)
    return avg_succ_prob, makespan


def wilson_score_interval(p_hat, n, z):
    denominator = 1 + z**2 / n
    centre_adjusted_probability = p_hat + z**2 / (2 * n)
    adjusted_standard_deviation = z * math.sqrt(
        (p_hat * (1 - p_hat) + z**2 / (4 * n)) / n
    )

    lower_bound = (
        centre_adjusted_probability - adjusted_standard_deviation
    ) / denominator
    upper_bound = (
        centre_adjusted_probability + adjusted_standard_deviation
    ) / denominator

    return (lower_bound, upper_bound)


@dataclass
class DataPoint:
    sweep_param: str
    sweep_value: float
    succ_prob: float
    succ_lo: float
    succ_hi: float
    makespan: float


@dataclass
class DataMeta:
    timestamp: str
    num_runs: int
    sim_duration: float
    num_clients: int
    t1: float
    t2: float
    sweep_param: str
    cc: Optional[float]
    sched_latency: Optional[float]
    bin_length: Optional[float]


@dataclass
class Data:
    meta: DataMeta
    data_points: List[DataPoint]


def sweep_cc(num_runs: int) -> Data:
    num_clients = 10
    t1 = 1e8
    t2 = 1e7
    sched_latency = 0
    bin_length = 1e5

    data_points: List[DataPoint] = []

    start = time.time()
    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    for cc in [1e5, 1e6, 1e7]:
        successes: List[float] = []
        makespans: List[float] = []

        for _ in range(num_runs):
            succ, makespan = bqc_trap(
                num_clients,
                1,
                t1=t1,
                t2=t2,
                cc=cc,
                sched_latency=sched_latency,
                bin_length=bin_length,
            )
            successes.append(succ)
            makespans.append(makespan)

        avg_succ = round(sum(successes) / len(successes), 3)
        total_makespan = sum(makespans)

        succ_lo, succ_hi = wilson_score_interval(
            p_hat=avg_succ, n=len(successes) * num_clients, z=1.96
        )
        succ_lo = round(succ_lo, 3)
        succ_hi = round(succ_hi, 3)

        print(
            f"CC = {cc}: succ = {avg_succ} ({succ_lo}, {succ_hi}), makespan = {total_makespan:_}"
        )

        data_point = DataPoint(
            sweep_param="cc",
            sweep_value=cc,
            succ_prob=avg_succ,
            succ_lo=succ_lo,
            succ_hi=succ_hi,
            makespan=total_makespan,
        )
        data_points.append(data_point)

    end = time.time()
    duration = round(end - start, 2)
    meta = DataMeta(
        timestamp=timestamp,
        num_runs=num_runs,
        sim_duration=duration,
        num_clients=num_clients,
        t1=t1,
        t2=t2,
        sweep_param="cc",
        cc=None,
        sched_latency=sched_latency,
        bin_length=bin_length,
    )
    return Data(meta, data_points)


def sweep_sched_latency(num_runs: int) -> Data:
    num_clients = 10
    t1 = 1e8
    t2 = 1e7
    cc = 1e5
    bin_length = 1e5

    data_points: List[DataPoint] = []

    start = time.time()
    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    for sched_latency in [1e3, 1e5, 1e7]:
        successes: List[float] = []
        makespans: List[float] = []

        for _ in range(num_runs):
            succ, makespan = bqc_trap(
                num_clients,
                1,
                t1=t1,
                t2=t2,
                cc=cc,
                sched_latency=sched_latency,
                bin_length=bin_length,
            )
            successes.append(succ)
            makespans.append(makespan)

        avg_succ = round(sum(successes) / len(successes), 3)
        total_makespan = sum(makespans)

        succ_lo, succ_hi = wilson_score_interval(
            p_hat=avg_succ, n=len(successes) * num_clients, z=1.96
        )
        succ_lo = round(succ_lo, 3)
        succ_hi = round(succ_hi, 3)

        print(
            f"Sched latency = {sched_latency}: succ = {avg_succ} ({succ_lo}, {succ_hi}), "
            f"makespan = {total_makespan:_}"
        )

        data_point = DataPoint(
            sweep_param="sched_latency",
            sweep_value=sched_latency,
            succ_prob=avg_succ,
            succ_lo=succ_lo,
            succ_hi=succ_hi,
            makespan=total_makespan,
        )
        data_points.append(data_point)

    end = time.time()
    duration = round(end - start, 2)
    meta = DataMeta(
        timestamp=timestamp,
        num_runs=num_runs,
        sim_duration=duration,
        num_clients=num_clients,
        t1=t1,
        t2=t2,
        sweep_param="sched_latency",
        cc=cc,
        sched_latency=None,
        bin_length=bin_length,
    )
    return Data(meta, data_points)


def sweep_bin_length(num_runs: int) -> Data:
    num_clients = 10
    t1 = 1e8
    t2 = 1e7
    cc = 1e5
    sched_latency = 0

    data_points: List[DataPoint] = []

    start = time.time()
    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    for bin_length in [1e5, 1e6, 1e7]:
        successes: List[float] = []
        makespans: List[float] = []

        for _ in range(num_runs):
            succ, makespan = bqc_trap(
                num_clients,
                1,
                t1=t1,
                t2=t2,
                cc=cc,
                sched_latency=sched_latency,
                bin_length=bin_length,
            )
            successes.append(succ)
            makespans.append(makespan)

        avg_succ = round(sum(successes) / len(successes), 3)
        total_makespan = sum(makespans)

        succ_lo, succ_hi = wilson_score_interval(
            p_hat=avg_succ, n=len(successes) * num_clients, z=1.96
        )
        succ_lo = round(succ_lo, 3)
        succ_hi = round(succ_hi, 3)

        print(
            f"Sched latency = {sched_latency}: succ = {avg_succ} ({succ_lo}, {succ_hi}), "
            f"makespan = {total_makespan:_}"
        )

        data_point = DataPoint(
            sweep_param="bin_length",
            sweep_value=bin_length,
            succ_prob=avg_succ,
            succ_lo=succ_lo,
            succ_hi=succ_hi,
            makespan=total_makespan,
        )
        data_points.append(data_point)

    end = time.time()
    duration = round(end - start, 2)
    meta = DataMeta(
        timestamp=timestamp,
        num_runs=num_runs,
        sim_duration=duration,
        num_clients=num_clients,
        t1=t1,
        t2=t2,
        sweep_param="bin_length",
        cc=cc,
        sched_latency=sched_latency,
        bin_length=None,
    )
    return Data(meta, data_points)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--num_runs", "-n", type=int, required=True)
    parser.add_argument("--sweep", "-s", type=str, required=True)

    args = parser.parse_args()
    num_runs = args.num_runs
    sweep = args.sweep

    data: Data
    output_dir: str

    if sweep == "cc":
        output_dir = "sweep_dir"
        data = sweep_cc(num_runs)
    elif sweep == "sched_latency":
        output_dir = "sweep_sched_latency"
        data = sweep_sched_latency(num_runs)
    elif sweep == "bin_length":
        output_dir = "sweep_bin_length"
        data = sweep_bin_length(num_runs)
    else:
        raise ValueError

    abs_dir = relative_to_cwd(f"data/{output_dir}")
    Path(abs_dir).mkdir(parents=True, exist_ok=True)
    last_path = os.path.join(abs_dir, "LAST.json")
    timestamp = data.meta.timestamp
    timestamp_path = os.path.join(abs_dir, f"{timestamp}.json")

    json_data = asdict(data)

    with open(last_path, "w") as datafile:
        json.dump(json_data, datafile)
    with open(timestamp_path, "w") as datafile:
        json.dump(json_data, datafile)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Dict, List, Optional, Union

from netqasm.lang.operand import Template

from qoala.lang.hostlang import IqoalaVector


class CallbackType(Enum):
    SEQUENTIAL = 0
    WAIT_ALL = auto()


class EprType(Enum):
    CREATE_KEEP = 0
    MEASURE_DIRECTLY = auto()
    REMOTE_STATE_PREP = auto()


class EprRole(Enum):
    CREATE = 0
    RECEIVE = auto()


class VirtIdMappingType(Enum):
    # All virt IDs have the same single value.
    # E.g. if single_value = 0 and num_pairs = 4,
    # virt IDs are [0, 0, 0, 0]
    EQUAL = 0

    # Virt IDs are increasing sequence starting at given value.
    # E.g. if single_value = 0 and num_pairs = 4,
    # virt IDs are [0, 1, 2, 3]
    INCREMENT = auto()

    # Explicit list of virt IDs used. Length needs to be equal to num_pairs.
    CUSTOM = auto()


@dataclass
class RequestVirtIdMapping:
    typ: VirtIdMappingType
    # Only allow templates with "all" or "increment", not with "custom"
    single_value: Optional[Union[Template, int]]
    custom_values: Optional[List[int]]

    def get_id(self, index: int) -> int:
        if self.typ == VirtIdMappingType.EQUAL:
            assert isinstance(self.single_value, int)
            return self.single_value
        elif self.typ == VirtIdMappingType.INCREMENT:
            assert isinstance(self.single_value, int)
            return self.single_value + index
        elif self.typ == VirtIdMappingType.CUSTOM:
            assert self.custom_values is not None
            return self.custom_values[index]
        raise ValueError

    def __str__(self) -> str:
        if self.typ == VirtIdMappingType.EQUAL:
            return f"all {self.single_value}"
        elif self.typ == VirtIdMappingType.INCREMENT:
            return f"increment {self.single_value}"
        elif self.typ == VirtIdMappingType.CUSTOM:
            assert self.custom_values is not None
            return f"custom {', '.join(str(v) for v in self.custom_values)}"
        raise ValueError

    @classmethod
    def from_str(cls, text: str) -> RequestVirtIdMapping:
        if text.startswith("all "):
            value_str = text[4:]
            if value_str.startswith("{") and value_str.endswith("}"):
                value_str = value_str.strip("{}").strip()
                value = Template(value_str)
            else:
                value = int(value_str)
            return RequestVirtIdMapping(
                typ=VirtIdMappingType.EQUAL, single_value=value, custom_values=None
            )
        elif text.startswith("increment "):
            value_str = text[10:]
            if value_str.startswith("{") and value_str.endswith("}"):
                value_str = value_str.strip("{}").strip()
                value = Template(value_str)
            else:
                value = int(value_str)
            return RequestVirtIdMapping(
                typ=VirtIdMappingType.INCREMENT, single_value=value, custom_values=None
            )
        elif text.startswith("custom "):
            int_list = text[7:]
            ints = [int(i) for i in int_list.split(", ")]
            return RequestVirtIdMapping(
                typ=VirtIdMappingType.CUSTOM, single_value=None, custom_values=ints
            )
        raise ValueError


@dataclass
class QoalaRequest:
    name: str  # TODO: remove?
    remote_id: Union[int, Template]
    epr_socket_id: Union[int, Template]
    num_pairs: Union[int, Template]
    virt_ids: RequestVirtIdMapping
    timeout: Union[float, Template]
    fidelity: Union[float, Template]
    typ: EprType
    role: EprRole

    def instantiate(self, values: Dict[str, Any]) -> None:
        if isinstance(self.remote_id, Template):
            self.remote_id = values[self.remote_id.name]
        if isinstance(self.epr_socket_id, Template):
            self.epr_socket_id = values[self.epr_socket_id.name]
        if isinstance(self.num_pairs, Template):
            self.num_pairs = values[self.num_pairs.name]
        if isinstance(self.virt_ids.single_value, Template):
            # Only need to check single_value. If "custom", singe_value is None,
            # and custom values themselves are never Templates.
            self.virt_ids.single_value = values[self.virt_ids.single_value.name]  # type: ignore
        if isinstance(self.timeout, Template):
            self.timeout = values[self.timeout.name]
        if isinstance(self.fidelity, Template):
            self.fidelity = values[self.fidelity.name]

    def serialize(self) -> str:
        s = f"REQUEST {self.name}"
        s += f"remote_id: {self.remote_id}"
        s += f"epr_socket_id: {self.epr_socket_id}"
        s += f"num_pairs: {self.num_pairs}"
        s += f"virt_ids: {self.virt_ids}"
        s += f"timeout: {self.timeout}"
        s += f"fidelity: {self.fidelity}"
        s += f"typ: {self.typ.name}"
        s += f"role: {self.role}"
        return s

    def __str__(self) -> str:
        return self.serialize()


@dataclass
class RequestRoutine:
    name: str
    request: QoalaRequest

    return_vars: List[Union[str, IqoalaVector]]

    callback_type: CallbackType
    callback: Optional[str]  # Local Routine name

    def instantiate(self, values: Dict[str, Any]) -> None:
        for i in range(len(self.return_vars)):
            ret_var = self.return_vars[i]
            if isinstance(ret_var, IqoalaVector):
                if isinstance(ret_var.size, Template):
                    size = values[ret_var.size.name]
                    # print(f"instantiating ret_var {ret_var.name} with size {size}")
                    self.return_vars[i] = IqoalaVector(ret_var.name, size)
        self.request.instantiate(values)

    def get_return_size(self, prog_input: Optional[Dict[str, int]] = None) -> int:
        size = 0
        for v in self.return_vars:
            if isinstance(v, IqoalaVector):
                if isinstance(v.size, int):
                    size += v.size
                elif isinstance(v.size, Template):
                    # Size is a template. It should be in the Program Input.
                    assert prog_input is not None
                    size += prog_input[v.size.name]
            else:
                size += 1
        return size



from __future__ import annotations

import logging
from copy import deepcopy
from typing import Any, Dict, Generator, Optional, Union

import netsquid as ns
from netqasm.lang.instr import NetQASMInstruction, core, nv, trapped_ion, vanilla
from netqasm.lang.operand import Register
from netsquid.components.instructions import (
    INSTR_CNOT,
    INSTR_CXDIR,
    INSTR_CYDIR,
    INSTR_CZ,
    INSTR_H,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_ROT_Y,
    INSTR_ROT_Z,
    INSTR_X,
    INSTR_Y,
    INSTR_Z,
)
from netsquid.components.instructions import Instruction as NsInstr

from pydynaa import EventExpression
from qoala.lang.routine import LocalRoutine
from qoala.runtime.instructions import (
    INSTR_BICHROMATIC,
    INSTR_MEASURE_ALL,
    INSTR_ROT_X_ALL,
    INSTR_ROT_Y_ALL,
    INSTR_ROT_Z_ALL,
)
from qoala.runtime.memory import ProgramMemory, RunningLocalRoutine
from qoala.runtime.sharedmem import MemAddr
from qoala.sim.memmgr import NotAllocatedError
from qoala.sim.process import QoalaProcess
from qoala.sim.qdevice import QDevice, QDeviceCommand
from qoala.sim.qnos.qnosinterface import QnosInterface, QnosLatencies
from qoala.util.logging import LogManager
from qoala.util.math import PI, PI_OVER_2


class UnsupportedNetqasmInstructionError(Exception):
    pass


class QnosProcessor:
    """Does not have state itself."""

    def __init__(
        self,
        interface: QnosInterface,
        latencies: QnosLatencies,
        asynchronous: bool = False,
    ) -> None:
        self._interface = interface
        self._latencies = latencies
        self._asynchronous = asynchronous

        # TODO: rewrite
        self._name = f"{interface.name}_QnosProcessor"

        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({self._name})"
        )

        # memory of current program, only not-None when processor is active
        self._current_prog_mem: Optional[ProgramMemory] = None

        self._current_routine: Optional[RunningLocalRoutine] = None

    def _prog_mem(self) -> ProgramMemory:
        # May only be called when processor is active
        assert self._current_prog_mem is not None
        return self._current_prog_mem

    def _routine(self) -> RunningLocalRoutine:
        # May only be called when processor is active
        assert self._current_routine is not None
        return self._current_routine

    @property
    def qdevice(self) -> QDevice:
        return self._interface.qdevice

    @property
    def interface(self) -> QnosInterface:
        return self._interface

    def instantiate_routine(
        self,
        process: QoalaProcess,
        routine: LocalRoutine,
        args: Dict[str, Any],
        input_addr: MemAddr,
        result_addr: MemAddr,
    ) -> None:
        """Instantiates and activates routine."""
        instance = deepcopy(routine)
        instance.subroutine.instantiate(process.pid, args)

        running_routine = RunningLocalRoutine(instance, input_addr, result_addr)
        process.qnos_mem.add_running_local_routine(running_routine)

    def assign_local_routine(
        self,
        process: QoalaProcess,
        routine_name: str,
        input_addr: MemAddr,
        result_addr: MemAddr,
    ) -> Generator[EventExpression, None, None]:
        routine = process.get_local_routine(routine_name)
        global_args = process.prog_instance.inputs.values

        uses = routine.metadata.qubit_use
        # Verify that all qubits that are used have been allocated.
        for q in uses:
            if not self._interface.memmgr.phys_id_for(process.pid, q) is not None:
                raise NotAllocatedError

        self.instantiate_routine(process, routine, global_args, input_addr, result_addr)

        netqasm_instrs = routine.subroutine.instructions

        instr_idx = 0
        while instr_idx < len(netqasm_instrs):
            instr_idx = yield from self.assign_routine_instr(
                process, routine_name, instr_idx
            )

    def assign_routine_instr(
        self, process: QoalaProcess, subrt_name: str, instr_idx: int
    ) -> Generator[EventExpression, None, int]:
        """Assign the processor to one specific instruction in a local routine."""
        running_routine = process.qnos_mem.get_running_local_routine(subrt_name)
        routine = running_routine.routine
        pid = process.prog_instance.pid

        self._current_prog_mem = process.prog_memory
        self._current_routine = running_routine

        subroutine = routine.subroutine

        # TODO: handle program counter and jumping!!
        instr = subroutine.instructions[instr_idx]
        self._logger.debug(
            f"{ns.sim_time()} interpreting instruction {instr_idx}: {instr}"
        )

        next_instr_idx: int

        if (
            isinstance(instr, core.JmpInstruction)
            or isinstance(instr, core.BranchUnaryInstruction)
            or isinstance(instr, core.BranchBinaryInstruction)
        ):
            if new_line_relative := (
                yield from self._interpret_branch_instr(pid, instr)
            ):
                next_instr_idx = instr_idx + new_line_relative
            else:
                next_instr_idx = instr_idx + 1
        else:
            generator = self._interpret_instruction(pid, instr)
            if generator:
                yield from generator
            next_instr_idx = instr_idx + 1

        self._current_prog_mem = None
        self._current_routine = None
        return next_instr_idx

    def _interpret_instruction(
        self, pid: int, instr: NetQASMInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        if isinstance(instr, core.SetInstruction):
            return self._interpret_set(pid, instr)
        elif isinstance(instr, core.StoreInstruction):
            return self._interpret_store(pid, instr)
        elif isinstance(instr, core.LoadInstruction):
            return self._interpret_load(pid, instr)
        elif isinstance(instr, core.LeaInstruction):
            return self._interpret_lea(pid, instr)
        elif isinstance(instr, core.InitInstruction):
            return self._interpret_init(pid, instr)
        elif isinstance(instr, core.MeasInstruction):
            return self._interpret_meas(pid, instr)
        elif isinstance(instr, core.SingleQubitInstruction):
            return self._interpret_single_qubit_instr(pid, instr)
        elif isinstance(instr, core.TwoQubitInstruction):
            return self._interpret_two_qubit_instr(pid, instr)
        elif isinstance(instr, core.RotationInstruction):
            return self._interpret_single_rotation_instr(pid, instr)
        elif isinstance(instr, core.ControlledRotationInstruction):
            return self._interpret_controlled_rotation_instr(pid, instr)
        elif isinstance(instr, core.ClassicalOpInstruction) or isinstance(
            instr, core.ClassicalOpModInstruction
        ):
            return self._interpret_binary_classical_instr(pid, instr)
        elif isinstance(instr, core.BreakpointInstruction):
            return self._interpret_breakpoint(pid, instr)
        elif isinstance(instr, NetQASMInstruction):
            return self._interpret_other_instr(pid, instr)
        else:
            raise RuntimeError(f"Invalid instruction {instr}")

    def _interpret_breakpoint(
        self, pid: int, instr: core.BreakpointInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        if instr.action.value == 0:
            self._logger.info("BREAKPOINT: no action taken")
            return None
        elif instr.action.value == 1:
            # TODO: fix
            raise NotImplementedError
            # self._logger.info("BREAKPOINT: dumping local state:")
            # for i in range(self.qdevice.qprocessor.num_positions):
            #     if self.qdevice.qprocessor.mem_positions[i].in_use:
            #         q = self.qdevice.qprocessor.peek(i)
            #         qstate = qubitapi.reduced_dm(q)
            #         self._logger.info(f"physical qubit {i}:\n{qstate}")

            # TODO: fix this; GlobalSimData is not static anymore!
            # state = GlobalSimData.get_quantum_state(save=True)  # type: ignore
        elif instr.action.value == 2:
            # TODO: fix
            raise NotImplementedError
            # self._logger.info("BREAKPOINT: dumping global state:")
            # if instr.role.value == 0:
            #     self._interface.send_netstack_msg(
            #         Message(content=NetstackBreakpointCreateRequest(pid))
            #     )
            #     ready = yield from self._interface.receive_netstack_msg()
            #     assert ready.content == "breakpoint ready"

            #     # TODO: fix this; GlobalSimData is not static anymore!
            #     state = GlobalSimData.get_quantum_state(save=True)  # type: ignore
            #     self._logger.info(state)

            #     self._interface.send_netstack_msg(Message(content="breakpoint end"))
            #     finished = yield from self._interface.receive_netstack_msg()
            #     assert finished.content == "breakpoint finished"
            # elif instr.role.value == 1:
            #     self._interface.send_netstack_msg(
            #         Message(content=NetstackBreakpointReceiveRequest(pid))
            #     )
            #     ready = yield from self._interface.receive_netstack_msg()
            #     assert ready.content == "breakpoint ready"
            #     self._interface.send_netstack_msg(Message(content="breakpoint end"))
            #     finished = yield from self._interface.receive_netstack_msg()
            #     assert finished.content == "breakpoint finished"
            # else:
            #     raise ValueError
        else:
            raise ValueError

    def _interpret_set(
        self, pid: int, instr: core.SetInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        self._logger.debug(f"Set register {instr.reg} to {instr.imm}")
        qnos_mem = self._prog_mem().qnos_mem

        # Simulate instruction duration.
        yield from self._interface.wait(self._latencies.qnos_instr_time)

        qnos_mem.set_reg_value(instr.reg, instr.imm.value)
        return None

    def _interpret_store(
        self, pid: int, instr: core.StoreInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        qnos_mem = self._prog_mem().qnos_mem

        shared_mem = self._prog_mem().shared_mem
        result_addr = self._routine().result_addr

        value = qnos_mem.get_reg_value(instr.reg)

        addr = instr.entry.address.address
        # Only allow NetQASM 2.0 input/output addresses
        assert isinstance(addr, str)
        assert addr == "output"
        entry = instr.entry.index
        assert isinstance(entry, Register)
        index = qnos_mem.get_reg_value(entry)
        self._logger.debug(
            f"Storing value {value} from register {instr.reg} "
            f"to array entry {instr.entry} (index {entry} = {index})"
        )

        # Simulate instruction duration.
        yield from self._interface.wait(self._latencies.qnos_instr_time)

        shared_mem.write_lr_out(result_addr, [value], offset=index)

        return None

    def _interpret_load(
        self, pid: int, instr: core.LoadInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        qnos_mem = self._prog_mem().qnos_mem

        shared_mem = self._prog_mem().shared_mem
        input_addr = self._routine().params_addr

        addr = instr.entry.address.address
        # Only allow NetQASM 2.0 input/output addresses
        assert isinstance(addr, str)
        assert addr == "input"
        entry = instr.entry.index
        assert isinstance(entry, Register)
        index = qnos_mem.get_reg_value(entry)
        [value] = shared_mem.read_lr_in(input_addr, 1, offset=index)

        if value is None:
            raise RuntimeError(f"array value at {instr.entry} is not defined")
        self._logger.debug(
            f"Storing value {value} from array entry {instr.entry} "
            f"to register {instr.reg}"
        )

        # Simulate instruction duration.
        yield from self._interface.wait(self._latencies.qnos_instr_time)

        qnos_mem.set_reg_value(instr.reg, value)
        return None

    def _interpret_lea(
        self, pid: int, instr: core.LeaInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        qnos_mem = self._prog_mem().qnos_mem
        self._logger.debug(
            f"Storing address of {instr.address} to register {instr.reg}"
        )

        # Simulate instruction duration.
        yield from self._interface.wait(self._latencies.qnos_instr_time)

        qnos_mem.set_reg_value(instr.reg, instr.address.address)
        return None

    def _interpret_branch_instr(
        self,
        pid: int,
        instr: Union[
            core.BranchUnaryInstruction,
            core.BranchBinaryInstruction,
            core.JmpInstruction,
        ],
    ) -> Generator[EventExpression, None, Optional[int]]:
        """Returns line to jump to, or None if no jump happens."""
        qnos_mem = self._prog_mem().qnos_mem
        a, b = None, None
        registers = []
        condition = False
        if isinstance(instr, core.BranchUnaryInstruction):
            a = qnos_mem.get_reg_value(instr.reg)
            registers = [instr.reg]
            condition = instr.check_condition(a)
        elif isinstance(instr, core.BranchBinaryInstruction):
            a = qnos_mem.get_reg_value(instr.reg0)
            b = qnos_mem.get_reg_value(instr.reg1)
            registers = [instr.reg0, instr.reg1]
            condition = instr.check_condition(a, b)
        elif isinstance(instr, core.JmpInstruction):
            condition = True

        # Simulate instruction duration.
        yield from self._interface.wait(self._latencies.qnos_instr_time)
        if condition:
            jump_address = instr.line
            self._logger.debug(
                f"Branching to line {jump_address}, since {instr}(a={a}, b={b}) "
                f"is True, with values from registers {registers}"
            )
            return jump_address.value  # type: ignore
        else:
            self._logger.debug(
                f"Don't branch, since {instr}(a={a}, b={b}) "
                f"is False, with values from registers {registers}"
            )
            return None

    def _interpret_binary_classical_instr(
        self,
        pid: int,
        instr: Union[
            core.ClassicalOpInstruction,
            core.ClassicalOpModInstruction,
        ],
    ) -> Optional[Generator[EventExpression, None, None]]:
        qnos_mem = self._prog_mem().qnos_mem
        mod = None
        if isinstance(instr, core.ClassicalOpModInstruction):
            mod = qnos_mem.get_reg_value(instr.regmod)
        if mod is not None and mod < 1:
            raise RuntimeError(f"Modulus needs to be greater or equal to 1, not {mod}")
        a = qnos_mem.get_reg_value(instr.regin0)
        b = qnos_mem.get_reg_value(instr.regin1)
        assert a is not None
        assert b is not None
        value = self._compute_binary_classical_instr(instr, a, b, mod=mod)
        mod_str = "" if mod is None else f"(mod {mod})"
        self._logger.debug(
            f"Performing {instr} of a={a} and b={b} {mod_str} "
            f"and storing the value {value} at register {instr.regout}"
        )

        # Simulate instruction duration.
        yield from self._interface.wait(self._latencies.qnos_instr_time)

        qnos_mem.set_reg_value(instr.regout, value)
        return None

    def _compute_binary_classical_instr(
        self, instr: NetQASMInstruction, a: int, b: int, mod: Optional[int] = 1
    ) -> int:
        if isinstance(instr, core.AddInstruction):
            return a + b
        elif isinstance(instr, core.AddmInstruction):
            assert mod is not None
            return (a + b) % mod
        elif isinstance(instr, core.SubInstruction):
            return a - b
        elif isinstance(instr, core.SubmInstruction):
            assert mod is not None
            return (a - b) % mod
        elif isinstance(instr, core.MulInstruction):
            return a * b
        elif isinstance(instr, core.DivInstruction):
            return a // b
        elif isinstance(instr, core.RemInstruction):
            return a % b
        else:
            raise ValueError(f"{instr} cannot be used as binary classical function")

    def _interpret_init(
        self, pid: int, instr: core.InitInstruction
    ) -> Generator[EventExpression, None, None]:
        raise NotImplementedError

    def _do_single_rotation(
        self,
        pid: int,
        instr: core.RotationInstruction,
        ns_instr: NsInstr,
    ) -> Generator[EventExpression, None, None]:
        qnos_mem = self._prog_mem().qnos_mem
        virt_id = qnos_mem.get_reg_value(instr.reg)
        phys_id = self._interface.memmgr.phys_id_for(pid, virt_id)
        if phys_id is None:
            raise NotAllocatedError
        if isinstance(instr.angle_num, Register):
            n = qnos_mem.get_reg_value(instr.angle_num)
        else:
            n = instr.angle_num.value
        if isinstance(instr.angle_denom, Register):
            d = qnos_mem.get_reg_value(instr.angle_denom)
        else:
            d = instr.angle_denom.value
        angle = self._get_rotation_angle_from_operands(n=n, d=d)
        self._logger.debug(
            f"Performing {instr} with angle {angle} on virtual qubit "
            f"{virt_id} (physical ID: {phys_id})"
        )
        commands = [QDeviceCommand(ns_instr, [phys_id], angle=angle)]
        yield from self.qdevice.execute_commands(commands)
        return None

    def _interpret_single_rotation_instr(
        self, pid: int, instr: core.RotationInstruction
    ) -> Generator[EventExpression, None, None]:
        raise NotImplementedError

    def _do_controlled_rotation(
        self,
        pid: int,
        instr: core.ControlledRotationInstruction,
        ns_instr: NsInstr,
    ) -> Generator[EventExpression, None, None]:
        qnos_mem = self._prog_mem().qnos_mem
        virt_id0 = qnos_mem.get_reg_value(instr.reg0)
        phys_id0 = self._interface.memmgr.phys_id_for(pid, virt_id0)
        if phys_id0 is None:
            raise NotAllocatedError
        virt_id1 = qnos_mem.get_reg_value(instr.reg1)
        phys_id1 = self._interface.memmgr.phys_id_for(pid, virt_id1)
        if phys_id1 is None:
            raise NotAllocatedError
        angle = self._get_rotation_angle_from_operands(
            n=instr.angle_num.value,
            d=instr.angle_denom.value,
        )
        self._logger.debug(
            f"Performing {instr} with angle {angle} on virtual qubits "
            f"{virt_id0} and {virt_id1} (physical IDs: {phys_id0} and {phys_id1})"
        )
        commands = [QDeviceCommand(ns_instr, [phys_id0, phys_id1], angle=angle)]
        yield from self.qdevice.execute_commands(commands)
        return None

    def _interpret_controlled_rotation_instr(
        self, pid: int, instr: core.ControlledRotationInstruction
    ) -> Generator[EventExpression, None, None]:
        raise NotImplementedError

    def _get_rotation_angle_from_operands(self, n: int, d: int) -> float:
        return float(n * PI / (2**d))

    def _interpret_meas(
        self, pid: int, instr: core.MeasInstruction
    ) -> Generator[EventExpression, None, None]:
        raise NotImplementedError

    def _interpret_single_qubit_instr(
        self, pid: int, instr: core.SingleQubitInstruction
    ) -> Generator[EventExpression, None, None]:
        raise NotImplementedError

    def _interpret_two_qubit_instr(
        self, pid: int, instr: core.TwoQubitInstruction
    ) -> Generator[EventExpression, None, None]:
        raise NotImplementedError

    def _interpret_other_instr(
        self, pid: int, instr: NetQASMInstruction
    ) -> Generator[EventExpression, None, None]:
        raise NotImplementedError


class GenericProcessor(QnosProcessor):
    """A `Processor` for nodes with a generic quantum hardware."""

    def _interpret_init(
        self, pid: int, instr: core.InitInstruction
    ) -> Generator[EventExpression, None, None]:
        qnos_mem = self._prog_mem().qnos_mem
        virt_id = qnos_mem.get_reg_value(instr.reg)
        phys_id = self._interface.memmgr.phys_id_for(pid, virt_id)
        if phys_id is None:
            raise NotAllocatedError
        self._logger.debug(
            f"Performing {instr} on virtual qubit "
            f"{virt_id} (physical ID: {phys_id})"
        )
        commands = [QDeviceCommand(INSTR_INIT, [phys_id])]
        yield from self.qdevice.execute_commands(commands)
        return None

    def _interpret_meas(
        self, pid: int, instr: core.MeasInstruction
    ) -> Generator[EventExpression, None, None]:
        qnos_mem = self._prog_mem().qnos_mem
        virt_id = qnos_mem.get_reg_value(instr.qreg)
        phys_id = self._interface.memmgr.phys_id_for(pid, virt_id)
        if phys_id is None:
            raise NotAllocatedError

        self._logger.debug(
            f"Measuring qubit {virt_id} (physical ID: {phys_id}), "
            f"placing the outcome in register {instr.creg}"
        )

        commands = [QDeviceCommand(INSTR_MEASURE, [phys_id])]
        outcome = yield from self.qdevice.execute_commands(commands)
        assert isinstance(outcome, int)
        qnos_mem.set_reg_value(instr.creg, outcome)
        return None

    def _interpret_single_qubit_instr(
        self, pid: int, instr: core.SingleQubitInstruction
    ) -> Generator[EventExpression, None, None]:
        qnos_mem = self._prog_mem().qnos_mem
        virt_id = qnos_mem.get_reg_value(instr.qreg)
        phys_id = self._interface.memmgr.phys_id_for(pid, virt_id)
        if phys_id is None:
            raise NotAllocatedError
        if isinstance(instr, vanilla.GateXInstruction):
            commands = [QDeviceCommand(INSTR_X, [phys_id])]
            yield from self.qdevice.execute_commands(commands)
        elif isinstance(instr, vanilla.GateYInstruction):
            commands = [QDeviceCommand(INSTR_Y, [phys_id])]
            yield from self.qdevice.execute_commands(commands)
        elif isinstance(instr, vanilla.GateZInstruction):
            commands = [QDeviceCommand(INSTR_Z, [phys_id])]
            yield from self.qdevice.execute_commands(commands)
        elif isinstance(instr, vanilla.GateHInstruction):
            commands = [QDeviceCommand(INSTR_H, [phys_id])]
            yield from self.qdevice.execute_commands(commands)
        else:
            raise UnsupportedNetqasmInstructionError
        return None

    def _interpret_single_rotation_instr(
        self, pid: int, instr: core.RotationInstruction
    ) -> Generator[EventExpression, None, None]:
        if isinstance(instr, vanilla.RotXInstruction):
            yield from self._do_single_rotation(pid, instr, INSTR_ROT_X)
        elif isinstance(instr, vanilla.RotYInstruction):
            yield from self._do_single_rotation(pid, instr, INSTR_ROT_Y)
        elif isinstance(instr, vanilla.RotZInstruction):
            yield from self._do_single_rotation(pid, instr, INSTR_ROT_Z)
        else:
            raise UnsupportedNetqasmInstructionError
        return None

    def _interpret_controlled_rotation_instr(
        self, pid: int, instr: core.ControlledRotationInstruction
    ) -> Generator[EventExpression, None, None]:
        raise UnsupportedNetqasmInstructionError

    def _interpret_two_qubit_instr(
        self, pid: int, instr: core.TwoQubitInstruction
    ) -> Generator[EventExpression, None, None]:
        qnos_mem = self._prog_mem().qnos_mem
        virt_id0 = qnos_mem.get_reg_value(instr.reg0)
        phys_id0 = self._interface.memmgr.phys_id_for(pid, virt_id0)
        if phys_id0 is None:
            raise NotAllocatedError
        virt_id1 = qnos_mem.get_reg_value(instr.reg1)
        phys_id1 = self._interface.memmgr.phys_id_for(pid, virt_id1)
        if phys_id1 is None:
            raise NotAllocatedError
        if isinstance(instr, vanilla.CnotInstruction):
            commands = [QDeviceCommand(INSTR_CNOT, [phys_id0, phys_id1])]
            yield from self.qdevice.execute_commands(commands)
        elif isinstance(instr, vanilla.CphaseInstruction):
            commands = [QDeviceCommand(INSTR_CZ, [phys_id0, phys_id1])]
            yield from self.qdevice.execute_commands(commands)
        else:
            raise UnsupportedNetqasmInstructionError
        return None


class NVProcessor(QnosProcessor):
    """A `Processor` for nodes with a NV hardware."""

    def _interpret_init(
        self, pid: int, instr: core.InitInstruction
    ) -> Generator[EventExpression, None, None]:
        memmgr = self._interface.memmgr
        qnos_mem = self._prog_mem().qnos_mem
        virt_id = qnos_mem.get_reg_value(instr.reg)
        phys_id = memmgr.phys_id_for(pid, virt_id)
        if phys_id is None:
            raise NotAllocatedError
        self._logger.debug(
            f"Performing {instr} on virtual qubit "
            f"{virt_id} (physical ID: {phys_id})"
        )
        commands = [QDeviceCommand(INSTR_INIT, [phys_id])]
        if phys_id != 0 and memmgr.virt_id_for(pid, 0) is not None:
            self._logger.info(
                f"Initialization of virt qubit {virt_id} (phys ID {phys_id}) will "
                "trash the existing state of phys qubit 0."
            )
            commands.append(QDeviceCommand(INSTR_INIT, [0]))
        yield from self.qdevice.execute_commands(commands)
        return None

    def _measure_electron(self) -> Generator[EventExpression, None, int]:
        commands = [QDeviceCommand(INSTR_MEASURE, [0])]
        outcome = yield from self.qdevice.execute_commands(commands)
        assert outcome is not None
        return outcome  # type: ignore

    def _move_carbon_to_electron_for_measure(
        self, carbon_id: int
    ) -> Generator[EventExpression, None, None]:
        commands = [
            QDeviceCommand(INSTR_INIT, [0]),
            QDeviceCommand(INSTR_ROT_Y, [0], angle=PI_OVER_2),
            QDeviceCommand(INSTR_CYDIR, [0, carbon_id], angle=-PI_OVER_2),
            QDeviceCommand(INSTR_ROT_X, [0], angle=-PI_OVER_2),
            QDeviceCommand(INSTR_CXDIR, [0, carbon_id], angle=PI_OVER_2),
            QDeviceCommand(INSTR_ROT_Y, [0], angle=-PI_OVER_2),
        ]
        yield from self.qdevice.execute_commands(commands)
        return None

    def _move_carbon_to_electron(
        self, carbon_id: int
    ) -> Generator[EventExpression, None, None]:
        # TODO: CHECK SEQUENCE OF GATES!!!
        commands = [
            QDeviceCommand(INSTR_INIT, [0]),
            QDeviceCommand(INSTR_ROT_Y, [0], angle=PI_OVER_2),
            QDeviceCommand(INSTR_CYDIR, [0, carbon_id], angle=-PI_OVER_2),
            QDeviceCommand(INSTR_ROT_X, [0], angle=-PI_OVER_2),
            QDeviceCommand(INSTR_CXDIR, [0, carbon_id], angle=PI_OVER_2),
            QDeviceCommand(INSTR_ROT_Y, [0], angle=-PI_OVER_2),
        ]
        yield from self.qdevice.execute_commands(commands)
        return None

    def _move_electron_to_carbon(
        self, carbon_id: int
    ) -> Generator[EventExpression, None, None]:
        commands = [
            QDeviceCommand(INSTR_INIT, [carbon_id]),
            QDeviceCommand(INSTR_ROT_Y, [0], angle=PI_OVER_2),
            QDeviceCommand(INSTR_CYDIR, [0, carbon_id], angle=-PI_OVER_2),
            QDeviceCommand(INSTR_ROT_X, [0], angle=-PI_OVER_2),
            QDeviceCommand(INSTR_CXDIR, [0, carbon_id], angle=PI_OVER_2),
        ]
        yield from self.qdevice.execute_commands(commands)
        return None

    def _interpret_meas(
        self, pid: int, instr: core.MeasInstruction
    ) -> Generator[EventExpression, None, None]:
        qnos_mem = self._prog_mem().qnos_mem
        virt_id = qnos_mem.get_reg_value(instr.qreg)
        phys_id = self._interface.memmgr.phys_id_for(pid, virt_id)
        if phys_id is None:
            raise NotAllocatedError

        # Only the electron (phys ID 0) can be measured.
        # Measuring any other physical qubit (i.e one of the carbons) requires
        # freeing up the electron and moving the target qubit to the electron first.

        self._logger.debug(
            f"Measuring qubit {virt_id} (physical ID: {phys_id}), "
            f"placing the outcome in register {instr.creg}"
        )

        if phys_id == 0:
            # Measuring a comm qubit. This can be done immediately.
            outcome = yield from self._measure_electron()
            qnos_mem.set_reg_value(instr.creg, outcome)
            return None
        # else:

        # We want to measure a mem qubit.
        # Move its state to the comm qubit.
        yield from self._move_carbon_to_electron_for_measure(phys_id)

        # Measure comm qubit (containing state of qubit to measure).
        outcome = yield from self._measure_electron()
        qnos_mem.set_reg_value(instr.creg, outcome)
        return None

    def _interpret_single_rotation_instr(
        self, pid: int, instr: core.RotationInstruction
    ) -> Generator[EventExpression, None, None]:
        if isinstance(instr, nv.RotXInstruction):
            yield from self._do_single_rotation(pid, instr, INSTR_ROT_X)
        elif isinstance(instr, nv.RotYInstruction):
            yield from self._do_single_rotation(pid, instr, INSTR_ROT_Y)
        elif isinstance(instr, nv.RotZInstruction):
            yield from self._do_single_rotation(pid, instr, INSTR_ROT_Z)
        else:
            raise UnsupportedNetqasmInstructionError

    def _interpret_controlled_rotation_instr(
        self, pid: int, instr: core.ControlledRotationInstruction
    ) -> Generator[EventExpression, None, None]:
        if isinstance(instr, nv.ControlledRotXInstruction):
            yield from self._do_controlled_rotation(pid, instr, INSTR_CXDIR)
        elif isinstance(instr, nv.ControlledRotYInstruction):
            yield from self._do_controlled_rotation(pid, instr, INSTR_CYDIR)
        else:
            raise UnsupportedNetqasmInstructionError


class IonTrapProcessor(QnosProcessor):
    """A `Processor` for nodes with a Ion Trap hardware."""

    def _interpret_init(
        self, pid: int, instr: core.InitInstruction
    ) -> Generator[EventExpression, None, None]:
        qnos_mem = self._prog_mem().qnos_mem
        virt_id = qnos_mem.get_reg_value(instr.reg)
        phys_id = self._interface.memmgr.phys_id_for(pid, virt_id)
        if phys_id is None:
            raise NotAllocatedError
        self._logger.debug(
            f"Performing {instr} on virtual qubit "
            f"{virt_id} (physical ID: {phys_id})"
        )
        commands = [QDeviceCommand(INSTR_INIT, [phys_id])]
        yield from self.qdevice.execute_commands(commands)
        return None

    def _interpret_meas(
        self, pid: int, instr: core.MeasInstruction
    ) -> Generator[EventExpression, None, None]:
        qnos_mem = self._prog_mem().qnos_mem
        virt_id = qnos_mem.get_reg_value(instr.qreg)
        phys_id = self._interface.memmgr.phys_id_for(pid, virt_id)
        if phys_id is None:
            raise NotAllocatedError

        self._logger.debug(
            f"Measuring qubit {virt_id} (physical ID: {phys_id}), "
            f"placing the outcome in register {instr.creg}"
        )

        commands = [QDeviceCommand(INSTR_MEASURE, [phys_id])]
        outcome = yield from self.qdevice.execute_commands(commands)
        assert isinstance(outcome, int)
        qnos_mem.set_reg_value(instr.creg, outcome)
        return None

    def _interpret_single_rotation_instr(
        self, pid: int, instr: core.RotationInstruction
    ) -> Generator[EventExpression, None, None]:
        if isinstance(instr, trapped_ion.RotZInstruction):
            yield from self._do_single_rotation(pid, instr, INSTR_ROT_Z)
        else:
            raise UnsupportedNetqasmInstructionError
        return None

    def _interpret_all_qubit_init(self):
        commands = [QDeviceCommand(INSTR_INIT)]
        yield from self.qdevice.execute_commands(commands)

    def _interpret_all_qubit_meas(self, instr: trapped_ion.AllQubitsMeasInstruction):
        qnos_mem = self._prog_mem().qnos_mem
        offset = qnos_mem.get_reg_value(instr.reg)
        result_addr = self._routine().result_addr
        shared_mem = self._prog_mem().shared_mem
        addr = instr.address.address
        # Only allow NetQASM 2.0 input/output addresses
        assert isinstance(addr, str)
        assert addr == "output"
        self._logger.debug("Measuring all qubits")
        commands = [QDeviceCommand(INSTR_MEASURE_ALL)]
        outcome = yield from self.qdevice.execute_commands(commands, parallel=True)
        assert isinstance(outcome, list)
        shared_mem.write_lr_out(result_addr, outcome, offset=offset)

        return None

    def _interpret_all_qubit_rotation(
        self, instr: trapped_ion.AllQubitsRotationInstruction, ns_instr: NsInstr
    ):
        qnos_mem = self._prog_mem().qnos_mem

        if isinstance(instr.angle_num, Register):
            n = qnos_mem.get_reg_value(instr.angle_num)
        else:
            n = instr.angle_num.value
        if isinstance(instr.angle_denom, Register):
            d = qnos_mem.get_reg_value(instr.angle_denom)
        else:
            d = instr.angle_denom.value
        angle = self._get_rotation_angle_from_operands(n=n, d=d)
        self._logger.debug(f"Performing {instr} with angle {angle} on all qubit")
        commands = [QDeviceCommand(ns_instr, angle=angle)]
        yield from self.qdevice.execute_commands(commands)
        return None

    def _interpret_bichromatic_instr(self, instr: trapped_ion.BichromaticInstruction):
        qnos_mem = self._prog_mem().qnos_mem

        if isinstance(instr.angle_num, Register):
            n = qnos_mem.get_reg_value(instr.angle_num)
        else:
            n = instr.angle_num.value
        if isinstance(instr.angle_denom, Register):
            d = qnos_mem.get_reg_value(instr.angle_denom)
        else:
            d = instr.angle_denom.value
        angle = self._get_rotation_angle_from_operands(n=n, d=d)
        self._logger.debug(f"Performing {instr} with angle {angle} on all qubit")

        commands = [QDeviceCommand(INSTR_BICHROMATIC, angle=angle)]
        yield from self.qdevice.execute_commands(commands)
        return None

    def _interpret_other_instr(
        self, pid: int, instr: NetQASMInstruction
    ) -> Generator[EventExpression, None, None]:
        if isinstance(instr, trapped_ion.BichromaticInstruction):
            yield from self._interpret_bichromatic_instr(instr)
        elif isinstance(instr, trapped_ion.AllQubitsInitInstruction):
            yield from self._interpret_all_qubit_init()
        elif isinstance(instr, trapped_ion.AllQubitsMeasInstruction):
            yield from self._interpret_all_qubit_meas(instr)
        elif isinstance(instr, trapped_ion.AllQubitsRotXInstruction):
            yield from self._interpret_all_qubit_rotation(instr, INSTR_ROT_X_ALL)
        elif isinstance(instr, trapped_ion.AllQubitsRotYInstruction):
            yield from self._interpret_all_qubit_rotation(instr, INSTR_ROT_Y_ALL)
        elif isinstance(instr, trapped_ion.AllQubitsRotZInstruction):
            yield from self._interpret_all_qubit_rotation(instr, INSTR_ROT_Z_ALL)
        else:
            raise UnsupportedNetqasmInstructionError



import pytest

from qoala.lang.ehi import EhiNetworkSchedule, EhiNetworkTimebin


def test_network_schedule():
    # 0:    (1, 0, 2, 0)
    # 100:  (1, 1, 2, 1)
    # 200:  (1, 2, 2, 2)
    # 1000: (1, 0, 2, 0)
    # 1100: (1, 1, 2, 1)
    # 1200: (1, 2, 2, 2)
    # 2000: (1, 0, 2, 0)
    # 2100: (1, 1, 2, 1)
    # 2200: (1, 2, 2, 2)
    # etc.

    node1 = 1
    node2 = 2

    def bin(pid1: int, pid2: int) -> EhiNetworkTimebin:
        return EhiNetworkTimebin(frozenset({node1, node2}), {node1: pid1, node2: pid2})

    pattern = [
        bin(0, 0),
        bin(1, 1),
        bin(2, 2),
    ]
    schedule = EhiNetworkSchedule(
        bin_length=100, first_bin=0, bin_pattern=pattern, repeat_period=1000
    )
    assert schedule.next_bin(0) == (0, bin(0, 0))
    assert schedule.next_bin(80) == (100, bin(1, 1))
    assert schedule.next_bin(100) == (100, bin(1, 1))
    assert schedule.next_bin(180) == (200, bin(2, 2))
    assert schedule.next_bin(200) == (200, bin(2, 2))
    assert schedule.next_bin(280) == (1000, bin(0, 0))
    assert schedule.next_bin(900) == (1000, bin(0, 0))
    assert schedule.next_bin(1000) == (1000, bin(0, 0))
    assert schedule.next_bin(1080) == (1100, bin(1, 1))

    assert schedule.next_specific_bin(0, bin(0, 0)) == 0
    assert schedule.next_specific_bin(0, bin(1, 1)) == 100
    assert schedule.next_specific_bin(0, bin(2, 2)) == 200
    assert schedule.next_specific_bin(10, bin(0, 0)) == 990
    assert schedule.next_specific_bin(10, bin(1, 1)) == 90
    assert schedule.next_specific_bin(10, bin(2, 2)) == 190
    assert schedule.next_specific_bin(600, bin(0, 0)) == 400
    assert schedule.next_specific_bin(600, bin(1, 1)) == 500
    with pytest.raises(ValueError):
        schedule.next_specific_bin(0, bin(1, 2))


if __name__ == "__main__":
    test_network_schedule()



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Generator, List, Optional, Type

import netsquid as ns
from netqasm.lang.parsing import parse_text_subroutine
from netsquid.components import QuantumProcessor
from netsquid.qubits import ketstates

from pydynaa import EventExpression
from qoala.lang.ehi import EhiNetworkInfo, UnitModule
from qoala.lang.hostlang import (
    AssignCValueOp,
    BasicBlock,
    BasicBlockType,
    ClassicalIqoalaOp,
    IqoalaSingleton,
    ReceiveCMsgOp,
    SendCMsgOp,
)
from qoala.lang.parse import LocalRoutineParser, QoalaParser
from qoala.lang.program import LocalRoutine, ProgramMeta, QoalaProgram
from qoala.lang.request import (
    CallbackType,
    EprRole,
    EprType,
    QoalaRequest,
    RequestRoutine,
    RequestVirtIdMapping,
)
from qoala.lang.routine import RoutineMetadata
from qoala.runtime.lhi import LhiLatencies, LhiLinkInfo, LhiTopology, LhiTopologyBuilder
from qoala.runtime.lhi_to_ehi import LhiConverter
from qoala.runtime.memory import ProgramMemory
from qoala.runtime.message import Message, RrCallTuple
from qoala.runtime.ntf import GenericNtf, NtfInterface
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.sim.build import build_qprocessor_from_topology
from qoala.sim.entdist.entdist import EntDist
from qoala.sim.entdist.entdistcomp import EntDistComponent
from qoala.sim.eprsocket import EprSocket
from qoala.sim.host.csocket import ClassicalSocket
from qoala.sim.host.hostinterface import HostInterface
from qoala.sim.memmgr import MemoryManager
from qoala.sim.netstack import NetstackInterface
from qoala.sim.process import QoalaProcess
from qoala.sim.procnode import ProcNode
from qoala.sim.qdevice import QDevice, QDeviceCommand
from qoala.sim.scheduler import NodeScheduler
from qoala.util.math import has_multi_state
from qoala.util.tests import netsquid_run

MOCK_MESSAGE = Message(src_pid=0, dst_pid=0, content=42)
MOCK_QNOS_RET_REG = "R0"
MOCK_QNOS_RET_VALUE = 7


@dataclass(frozen=True)
class InterfaceEvent:
    peer: str
    msg: Message


@dataclass(frozen=True)
class FlushEvent:
    pass


@dataclass(frozen=True)
class SignalEvent:
    pass


class MockNetstackInterface(NetstackInterface):
    def __init__(
        self,
        qdevice: QDevice,
        memmgr: MemoryManager,
    ) -> None:
        self._qdevice = qdevice
        self._memmgr = memmgr

    def send_qnos_msg(self, msg: Message) -> None:
        return None

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        return None

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        return None
        yield  # to make this behave as a generator

    def send_entdist_msg(self, msg: Message) -> None:
        return None

    def receive_entdist_msg(self) -> Generator[EventExpression, None, Message]:
        return Message(-1, -1, 0)
        yield  # to make this behave as a generator

    @property
    def node_id(self) -> int:
        return 0


class MockQDevice(QDevice):
    def __init__(self, topology: LhiTopology) -> None:
        self._topology = topology

        self._executed_commands: List[QDeviceCommand] = []

    def set_mem_pos_in_use(self, id: int, in_use: bool) -> None:
        pass

    def execute_commands(
        self, commands: List[QDeviceCommand]
    ) -> Generator[EventExpression, None, Optional[int]]:
        self._executed_commands.extend(commands)
        return None
        yield

    def reset(self) -> None:
        self._executed_commands = []


class MockHostInterface(HostInterface):
    def __init__(self) -> None:
        self.send_events: List[InterfaceEvent] = []
        self.recv_events: List[InterfaceEvent] = []
        self._program_instance_jumps: Dict[int, int] = {}  # pid => block name

    @property
    def program_instance_jumps(self) -> Dict[int, int]:
        return self._program_instance_jumps

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        self.send_events.append(InterfaceEvent(peer, msg))

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent(peer, MOCK_MESSAGE))
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    def send_qnos_msg(self, msg: Message) -> None:
        self.send_events.append(InterfaceEvent("qnos", msg))

    def receive_qnos_msg(self) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent("qnos", MOCK_MESSAGE))
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    @property
    def name(self) -> str:
        return "mock"


def create_program(
    instrs: Optional[List[ClassicalIqoalaOp]] = None,
    subroutines: Optional[Dict[str, LocalRoutine]] = None,
    req_routines: Optional[Dict[str, RequestRoutine]] = None,
    meta: Optional[ProgramMeta] = None,
) -> QoalaProgram:
    if instrs is None:
        instrs = []
    if subroutines is None:
        subroutines = {}

    if req_routines is None:
        req_routines = {}
    if meta is None:
        meta = ProgramMeta.empty("prog")
    # TODO: split into proper blocks
    block = BasicBlock("b0", BasicBlockType.CL, instrs)
    return QoalaProgram(
        blocks=[block],
        local_routines=subroutines,
        meta=meta,
        request_routines=req_routines,
    )


def create_process(
    pid: int,
    program: QoalaProgram,
    unit_module: UnitModule,
    host_interface: HostInterface,
    inputs: Optional[Dict[str, Any]] = None,
) -> QoalaProcess:
    if inputs is None:
        inputs = {}
    prog_input = ProgramInput(values=inputs)
    instance = ProgramInstance(
        pid=pid,
        program=program,
        inputs=prog_input,
        unit_module=unit_module,
    )
    mem = ProgramMemory(pid=0)

    process = QoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={
            id: ClassicalSocket(host_interface, name, 0, 0)
            for (id, name) in program.meta.csockets.items()
        },
        epr_sockets={0: EprSocket(0, 0, 0, 0, 1.0)},
        result=ProgramResult(values={}),
    )
    return process


class MockScheduler(NodeScheduler):
    def __init__(self, scheduler: NodeScheduler) -> None:
        self._cpu_scheduler = scheduler.cpu_scheduler
        self._qpu_scheduler = scheduler.qpu_scheduler
        self._host = scheduler.host
        self._last_cpu_task_pid = -1
        self._last_qpu_task_pid = -1
        self._is_predictable = True
        pass

    def schedule_next_for(self, pid: int) -> None:
        pass

    def schedule_all(self) -> None:
        pass


def create_qprocessor(name: str, num_qubits: int) -> QuantumProcessor:
    topology = LhiTopologyBuilder.perfect_uniform_default_gates(num_qubits=num_qubits)
    return build_qprocessor_from_topology(name=f"{name}_processor", topology=topology)


def create_network_ehi(num_qubits: int, names: List[str] = None) -> EhiNetworkInfo:
    if names is None:
        names = ["alice", "bob", "charlie"]

    nodes = {i: name for i, name in enumerate(names)}
    return EhiNetworkInfo(nodes, links={})


def create_procnode(
    name: str,
    network_ehi: EhiNetworkInfo,
    num_qubits: int,
    topology: LhiTopology,
    latencies: LhiLatencies,
    ntf_interface: NtfInterface,
    procnode_cls: Type[ProcNode] = ProcNode,
    asynchronous: bool = False,
) -> ProcNode:
    alice_qprocessor = create_qprocessor(name, num_qubits)

    node_id = network_ehi.get_node_id(name)
    procnode = procnode_cls(
        name=name,
        qprocessor=alice_qprocessor,
        qdevice_topology=topology,
        latencies=latencies,
        ntf_interface=ntf_interface,
        network_ehi=network_ehi,
        node_id=node_id,
        asynchronous=asynchronous,
    )

    return procnode


def simple_subroutine(name: str, subrt_text: str) -> LocalRoutine:
    subrt = parse_text_subroutine(subrt_text)
    return LocalRoutine(
        name, subrt, return_vars=[], metadata=RoutineMetadata.use_none()
    )


def parse_iqoala_subroutines(subrt_text: str) -> LocalRoutine:
    return LocalRoutineParser(subrt_text).parse()


def generic_topology(num_qubits: int) -> LhiTopology:
    # Instructions and durations are not needed for these tests.
    return LhiTopologyBuilder.perfect_uniform(
        num_qubits=num_qubits,
        single_instructions=[],
        single_duration=0,
        two_instructions=[],
        two_duration=0,
    )


def star_topology(num_qubits: int) -> LhiTopology:
    # Instructions and durations are not needed for these tests.
    return LhiTopologyBuilder.perfect_star(
        num_qubits=num_qubits,
        comm_instructions=[],
        comm_duration=0,
        mem_instructions=[],
        mem_duration=0,
        two_instructions=[],
        two_duration=0,
    )


def test_initialize():
    num_qubits = 3
    topology = generic_topology(num_qubits)
    latencies = LhiLatencies.all_zero()
    ntf = GenericNtf()

    network_ehi = create_network_ehi(num_qubits)
    procnode = create_procnode(
        "alice", network_ehi, num_qubits, topology, latencies, ntf
    )
    procnode.qdevice = MockQDevice(topology)

    procnode.host.interface = MockHostInterface()

    procnode.netstack.interface = MockNetstackInterface(
        procnode.qdevice, procnode.memmgr
    )

    host_processor = procnode.host.processor
    qnos_processor = procnode.qnos.processor
    netstack_processor = procnode.netstack.processor

    ehi = LhiConverter.to_ehi(topology, ntf, latencies)
    unit_module = UnitModule.from_full_ehi(ehi)

    instrs = [AssignCValueOp(IqoalaSingleton("x"), 3)]
    subrt1 = simple_subroutine(
        "subrt1",
        """
    set R5 42
    """,
    )

    request_routine = RequestRoutine(
        name="req1",
        return_vars=[],
        callback_type=CallbackType.WAIT_ALL,
        callback=None,
        request=QoalaRequest(
            name="req1",
            remote_id=1,
            epr_socket_id=0,
            num_pairs=1,
            virt_ids=RequestVirtIdMapping.from_str("increment 0"),
            timeout=1000,
            fidelity=1.0,
            typ=EprType.CREATE_KEEP,
            role=EprRole.CREATE,
        ),
    )

    program = create_program(
        instrs=instrs,
        subroutines={"subrt1": subrt1},
        req_routines={"req1": request_routine},
    )
    process = create_process(
        pid=0,
        program=program,
        unit_module=unit_module,
        host_interface=procnode.host._interface,
        inputs={"x": 1, "theta": 3.14, "name": "alice"},
    )
    procnode.add_process(process)

    host_processor.initialize(process)
    host_mem = process.prog_memory.host_mem
    assert host_mem.read("x") == 1
    assert host_mem.read("theta") == 3.14
    assert host_mem.read("name") == "alice"

    netsquid_run(host_processor.assign_instr_index(process, instr_idx=0))
    routine = process.program.local_routines["subrt1"]
    qnos_processor.instantiate_routine(process, routine, {}, 0, 0)
    netsquid_run(qnos_processor.assign_routine_instr(process, "subrt1", 0))

    rrcall = RrCallTuple.no_alloc("req1")
    global_args = process.prog_instance.inputs.values
    netstack_processor.instantiate_routine(process, rrcall, global_args)
    netsquid_run(
        netstack_processor.handle_single_pair(
            process, "req1", 0
        )  # not s  ure about the 0
    )

    assert process.host_mem.read("x") == 3
    assert process.qnos_mem.get_reg_value("R5") == 42
    assert procnode.memmgr.phys_id_for(pid=process.pid, virt_id=0) == 0


def test_classical_comm():
    ns.sim_reset()

    num_qubits = 3

    topology = generic_topology(num_qubits)
    latencies = LhiLatencies.all_zero()
    ntf = GenericNtf()

    network_ehi = create_network_ehi(num_qubits)

    class TestProcNode(ProcNode):
        def run(self) -> Generator[EventExpression, None, None]:
            process = self.memmgr.get_process(0)
            yield from self.host.processor.assign_instr_index(process, 0)

    alice_procnode = create_procnode(
        "alice",
        network_ehi,
        num_qubits,
        topology,
        latencies,
        ntf,
        procnode_cls=TestProcNode,
    )
    bob_procnode = create_procnode(
        "bob",
        network_ehi,
        num_qubits,
        topology,
        latencies,
        ntf,
        procnode_cls=TestProcNode,
    )

    alice_host_processor = alice_procnode.host.processor
    bob_host_processor = bob_procnode.host.processor

    ehi = LhiConverter.to_ehi(topology, ntf, latencies)
    unit_module = UnitModule.from_full_ehi(ehi)

    alice_instrs = [
        SendCMsgOp(IqoalaSingleton("csocket_id"), IqoalaSingleton("message"))
    ]
    alice_meta = ProgramMeta(
        name="alice",
        parameters=["csocket_id", "message"],
        csockets={0: "bob"},
        epr_sockets={},
    )
    alice_program = create_program(instrs=alice_instrs, meta=alice_meta)
    alice_process = create_process(
        pid=0,
        program=alice_program,
        unit_module=unit_module,
        host_interface=alice_procnode.host._interface,
        inputs={"csocket_id": 0, "message": 1337},
    )
    alice_procnode.add_process(alice_process)
    alice_host_processor.initialize(alice_process)

    bob_instrs = [
        ReceiveCMsgOp(IqoalaSingleton("csocket_id"), IqoalaSingleton("result"))
    ]
    bob_meta = ProgramMeta(
        name="bob", parameters=["csocket_id"], csockets={0: "alice"}, epr_sockets={}
    )
    bob_program = create_program(instrs=bob_instrs, meta=bob_meta)
    bob_process = create_process(
        pid=0,
        program=bob_program,
        unit_module=unit_module,
        host_interface=bob_procnode.host._interface,
        inputs={"csocket_id": 0},
    )
    bob_procnode.add_process(bob_process)
    bob_host_processor.initialize(bob_process)

    alice_procnode.connect_to(bob_procnode)

    # To prevent scheduler from running
    alice_procnode.scheduler = MockScheduler(alice_procnode.scheduler)
    bob_procnode.scheduler = MockScheduler(bob_procnode.scheduler)

    # First start Bob, since Alice won't yield on anything (she only does a Send
    # instruction) and therefore calling 'start()' on alice completes her whole
    # protocol while Bob's interface has not even been started.

    bob_procnode.start()
    alice_procnode.start()

    ns.sim_run()

    assert bob_process.host_mem.read("result") == 1337


def test_classical_comm_three_nodes():
    ns.sim_reset()

    num_qubits = 3

    topology = generic_topology(num_qubits)
    latencies = LhiLatencies.all_zero()
    ntf = GenericNtf()

    network_ehi = create_network_ehi(num_qubits)

    class SenderProcNode(ProcNode):
        def run(self) -> Generator[EventExpression, None, None]:
            process = self.memmgr.get_process(0)
            yield from self.host.processor.assign_instr_index(process, 0)

    class ReceiverProcNode(ProcNode):
        def run(self) -> Generator[EventExpression, None, None]:
            process = self.memmgr.get_process(0)
            yield from self.host.processor.assign_instr_index(process, 0)
            yield from self.host.processor.assign_instr_index(process, 1)

    alice_procnode = create_procnode(
        "alice",
        network_ehi,
        num_qubits,
        topology,
        latencies,
        ntf,
        procnode_cls=SenderProcNode,
    )
    bob_procnode = create_procnode(
        "bob",
        network_ehi,
        num_qubits,
        topology,
        latencies,
        ntf,
        procnode_cls=SenderProcNode,
    )
    charlie_procnode = create_procnode(
        "charlie",
        network_ehi,
        num_qubits,
        topology,
        latencies,
        ntf,
        procnode_cls=ReceiverProcNode,
    )

    alice_host_processor = alice_procnode.host.processor
    bob_host_processor = bob_procnode.host.processor
    charlie_host_processor = charlie_procnode.host.processor

    ehi = LhiConverter.to_ehi(topology, ntf, latencies)
    unit_module = UnitModule.from_full_ehi(ehi)

    alice_instrs = [
        SendCMsgOp(IqoalaSingleton("csocket_id"), IqoalaSingleton("message"))
    ]
    alice_meta = ProgramMeta(
        name="alice",
        parameters=["csocket_id", "message"],
        csockets={0: "charlie"},
        epr_sockets={},
    )
    alice_program = create_program(instrs=alice_instrs, meta=alice_meta)
    alice_process = create_process(
        pid=0,
        program=alice_program,
        unit_module=unit_module,
        host_interface=alice_procnode.host._interface,
        inputs={"csocket_id": 0, "message": 1337},
    )
    alice_procnode.add_process(alice_process)
    alice_host_processor.initialize(alice_process)

    bob_instrs = [SendCMsgOp(IqoalaSingleton("csocket_id"), IqoalaSingleton("message"))]
    bob_meta = ProgramMeta(
        name="bob",
        parameters=["csocket_id", "message"],
        csockets={0: "charlie"},
        epr_sockets={},
    )
    bob_program = create_program(instrs=bob_instrs, meta=bob_meta)
    bob_process = create_process(
        pid=0,
        program=bob_program,
        unit_module=unit_module,
        host_interface=bob_procnode.host._interface,
        inputs={"csocket_id": 0, "message": 42},
    )
    bob_procnode.add_process(bob_process)
    bob_host_processor.initialize(bob_process)

    charlie_instrs = [
        ReceiveCMsgOp(
            IqoalaSingleton("csocket_id_alice"), IqoalaSingleton("result_alice")
        ),
        ReceiveCMsgOp(IqoalaSingleton("csocket_id_bob"), IqoalaSingleton("result_bob")),
    ]
    charlie_meta = ProgramMeta(
        name="bob",
        parameters=["csocket_id_alice", "csocket_id_bob"],
        csockets={0: "alice", 1: "bob"},
        epr_sockets={},
    )
    charlie_program = create_program(instrs=charlie_instrs, meta=charlie_meta)
    charlie_process = create_process(
        pid=0,
        program=charlie_program,
        unit_module=unit_module,
        host_interface=charlie_procnode.host._interface,
        inputs={"csocket_id_alice": 0, "csocket_id_bob": 1},
    )
    charlie_procnode.add_process(charlie_process)
    charlie_host_processor.initialize(charlie_process)

    alice_procnode.connect_to(charlie_procnode)
    bob_procnode.connect_to(charlie_procnode)

    # To prevent scheduler from running
    alice_procnode.scheduler = MockScheduler(alice_procnode.scheduler)
    bob_procnode.scheduler = MockScheduler(bob_procnode.scheduler)
    charlie_procnode.scheduler = MockScheduler(charlie_procnode.scheduler)

    # First start Charlie, since Alice and Bob don't yield on anything.
    charlie_procnode.start()
    alice_procnode.start()
    bob_procnode.start()
    ns.sim_run()

    assert charlie_process.host_mem.read("result_alice") == 1337
    assert charlie_process.host_mem.read("result_bob") == 42


def test_epr():
    ns.sim_reset()

    num_qubits = 3

    topology = generic_topology(num_qubits)
    latencies = LhiLatencies.all_zero()
    ntf = GenericNtf()

    network_ehi = create_network_ehi(num_qubits)
    alice_id = network_ehi.get_node_id("alice")
    bob_id = network_ehi.get_node_id("bob")

    class TestProcNode(ProcNode):
        def run(self) -> Generator[EventExpression, None, None]:
            process = self.memmgr.get_process(0)
            rrcall = RrCallTuple.no_alloc("req1")
            global_args = process.prog_instance.inputs.values
            self.netstack.processor.instantiate_routine(process, rrcall, global_args)
            yield from self.netstack.processor.handle_single_pair(
                process, "req1", 0
            )  # not s  ure about the 0

    alice_procnode = create_procnode(
        "alice",
        network_ehi,
        num_qubits,
        topology,
        latencies,
        ntf,
        procnode_cls=TestProcNode,
    )
    bob_procnode = create_procnode(
        "bob",
        network_ehi,
        num_qubits,
        topology,
        latencies,
        ntf,
        procnode_cls=TestProcNode,
    )

    alice_host_processor = alice_procnode.host.processor
    bob_host_processor = bob_procnode.host.processor

    ehi = LhiConverter.to_ehi(topology, ntf, latencies)
    unit_module = UnitModule.from_full_ehi(ehi)

    alice_request_routine = RequestRoutine(
        name="req1",
        return_vars=[],
        callback_type=CallbackType.WAIT_ALL,
        callback=None,
        request=QoalaRequest(
            name="req1",
            remote_id=bob_id,
            epr_socket_id=0,
            num_pairs=1,
            virt_ids=RequestVirtIdMapping.from_str("increment 0"),
            timeout=1000,
            fidelity=1.0,
            typ=EprType.CREATE_KEEP,
            role=EprRole.CREATE,
        ),
    )

    bob_request_routine = RequestRoutine(
        name="req1",
        return_vars=[],
        callback_type=CallbackType.WAIT_ALL,
        callback=None,
        request=QoalaRequest(
            name="req1",
            remote_id=alice_id,
            epr_socket_id=0,
            num_pairs=1,
            virt_ids=RequestVirtIdMapping.from_str("increment 0"),
            timeout=1000,
            fidelity=1.0,
            typ=EprType.CREATE_KEEP,
            role=EprRole.RECEIVE,
        ),
    )

    alice_instrs = [SendCMsgOp("csocket_id", "message")]
    alice_meta = ProgramMeta(
        name="alice",
        parameters=["csocket_id", "message"],
        csockets={0: "bob"},
        epr_sockets={},
    )
    alice_program = create_program(
        instrs=alice_instrs,
        req_routines={"req1": alice_request_routine},
        meta=alice_meta,
    )
    alice_process = create_process(
        pid=0,
        program=alice_program,
        unit_module=unit_module,
        host_interface=alice_procnode.host._interface,
        inputs={"csocket_id": 0, "message": 1337},
    )
    alice_procnode.add_process(alice_process)
    alice_host_processor.initialize(alice_process)

    bob_instrs = [ReceiveCMsgOp("csocket_id", "result")]
    bob_meta = ProgramMeta(
        name="bob", parameters=["csocket_id"], csockets={0: "alice"}, epr_sockets={}
    )
    bob_program = create_program(
        instrs=bob_instrs, req_routines={"req1": bob_request_routine}, meta=bob_meta
    )
    bob_process = create_process(
        pid=0,
        program=bob_program,
        unit_module=unit_module,
        host_interface=bob_procnode.host._interface,
        inputs={"csocket_id": 0},
    )
    bob_procnode.add_process(bob_process)
    bob_host_processor.initialize(bob_process)

    alice_procnode.connect_to(bob_procnode)

    nodes = [alice_procnode.node, bob_procnode.node]
    entdistcomp = EntDistComponent(network_ehi)
    alice_procnode.node.entdist_out_port.connect(entdistcomp.node_in_port("alice"))
    alice_procnode.node.entdist_in_port.connect(entdistcomp.node_out_port("alice"))
    bob_procnode.node.entdist_out_port.connect(entdistcomp.node_in_port("bob"))
    bob_procnode.node.entdist_in_port.connect(entdistcomp.node_out_port("bob"))
    entdist = EntDist(nodes=nodes, ehi_network=network_ehi, comp=entdistcomp)
    link_info = LhiLinkInfo.perfect(1000)
    entdist.add_sampler(alice_procnode.node.ID, bob_procnode.node.ID, link_info)

    # To prevent scheduler from running
    alice_procnode.scheduler = MockScheduler(alice_procnode.scheduler)
    bob_procnode.scheduler = MockScheduler(bob_procnode.scheduler)

    # First start Bob, since Alice won't yield on anything (she only does a Send
    # instruction) and therefore calling 'start()' on alice completes her whole
    # protocol while Bob's interface has not even been started.
    bob_procnode.start()
    alice_procnode.start()
    entdist.start()
    ns.sim_run()

    assert alice_procnode.memmgr.phys_id_for(pid=0, virt_id=0) == 0
    assert bob_procnode.memmgr.phys_id_for(pid=0, virt_id=0) == 0

    alice_qubit = alice_procnode.qdevice.get_local_qubit(0)
    bob_qubit = bob_procnode.qdevice.get_local_qubit(0)
    assert has_multi_state([alice_qubit, bob_qubit], ketstates.b00)


def test_whole_program():
    ns.sim_reset()

    server_text = """
META_START
    name: server
    parameters: client_id
    csockets: 0 -> client
    epr_sockets: 0 -> client
META_END

^b0 {type = CL}:
    remote_id = assign_cval() : {client_id}
^b1 {type = QC}:
    run_request() : req1

REQUEST req1
  callback_type: wait_all
  callback:
  return_vars: 
  remote_id: {client_id}
  epr_socket_id: 0
  num_pairs: 1
  virt_ids: all 0
  timeout:1000
  fidelity: 1.0
  typ: create_keep
  role: receive
    """
    server_program = QoalaParser(server_text).parse()

    num_qubits = 3
    topology = generic_topology(num_qubits)
    latencies = LhiLatencies.all_zero()
    ntf = GenericNtf()
    ehi = LhiConverter.to_ehi(topology, ntf, latencies)
    unit_module = UnitModule.from_full_ehi(ehi)
    network_ehi = create_network_ehi(num_qubits, names=["client", "server"])
    server_id = network_ehi.get_node_id("server")
    client_id = network_ehi.get_node_id("client")

    class TestProcNode(ProcNode):
        def run(self) -> Generator[EventExpression, None, None]:
            process = self.memmgr.get_process(0)
            self.scheduler.initialize_process(process)
            rrcall = RrCallTuple.no_alloc("req1")
            global_args = process.prog_instance.inputs.values
            self.netstack.processor.instantiate_routine(process, rrcall, global_args)
            yield from self.netstack.processor.handle_single_pair(
                process, "req1", 0
            )  # not s  ure about the 0

    server_procnode = create_procnode(
        "server",
        network_ehi,
        num_qubits,
        topology,
        latencies,
        ntf,
        procnode_cls=TestProcNode,
    )
    server_process = create_process(
        pid=0,
        program=server_program,
        unit_module=unit_module,
        host_interface=server_procnode.host._interface,
        inputs={"client_id": client_id, "csocket_id": 0, "message": 1337},
    )
    server_procnode.add_process(server_process)

    client_text = """
META_START
    name: client
    parameters: server_id
    csockets: 0 -> server
    epr_sockets: 0 -> server
META_END

^b0 {type = CL}:
    remote_id = assign_cval() : {server_id}
^b1 {type = QL}:
    run_request() : req1

REQUEST req1
  callback_type: wait_all
  callback:
  return_vars: 
  remote_id: {server_id}
  epr_socket_id: 0
  num_pairs: 1
  virt_ids: all 0
  timeout: 1000
  fidelity: 1.0
  typ: create_keep
  role: create
    """
    client_program = QoalaParser(client_text).parse()
    client_procnode = create_procnode(
        "client",
        network_ehi,
        num_qubits,
        topology,
        latencies,
        ntf,
        procnode_cls=TestProcNode,
    )
    client_process = create_process(
        pid=0,
        program=client_program,
        unit_module=unit_module,
        host_interface=client_procnode.host._interface,
        inputs={"server_id": server_id, "csocket_id": 0, "message": 1337},
    )
    client_procnode.add_process(client_process)

    client_procnode.connect_to(server_procnode)

    nodes = [client_procnode.node, server_procnode.node]
    entdistcomp = EntDistComponent(network_ehi)
    client_procnode.node.entdist_out_port.connect(entdistcomp.node_in_port("client"))
    client_procnode.node.entdist_in_port.connect(entdistcomp.node_out_port("client"))
    server_procnode.node.entdist_out_port.connect(entdistcomp.node_in_port("server"))
    server_procnode.node.entdist_in_port.connect(entdistcomp.node_out_port("server"))
    entdist = EntDist(nodes=nodes, ehi_network=network_ehi, comp=entdistcomp)
    link_info = LhiLinkInfo.perfect(1000)
    entdist.add_sampler(client_procnode.node.ID, server_procnode.node.ID, link_info)

    # To prevent scheduler from running
    server_procnode.scheduler = MockScheduler(server_procnode.scheduler)
    client_procnode.scheduler = MockScheduler(client_procnode.scheduler)

    server_procnode.start()
    client_procnode.start()
    entdist.start()
    ns.sim_run()

    assert client_procnode.memmgr.phys_id_for(pid=0, virt_id=0) == 0
    assert server_procnode.memmgr.phys_id_for(pid=0, virt_id=0) == 0
    client_qubit = client_procnode.qdevice.get_local_qubit(0)
    server_qubit = server_procnode.qdevice.get_local_qubit(0)
    assert has_multi_state([client_qubit, server_qubit], ketstates.b00)

    # assert client_procnode.memmgr.phys_id_for(pid=0, virt_id=1) == 1
    # assert server_procnode.memmgr.phys_id_for(pid=0, virt_id=1) == 1


if __name__ == "__main__":
    test_initialize()
    test_classical_comm()
    test_classical_comm_three_nodes()
    test_epr()
    test_whole_program()
